//# Array.h: A templated N-D Array class with variable origin
//# Copyright (C) 1993,1994
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: aips2-request@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id: Array.h,v 1.1.1.1 1996/04/04 16:37:03 dschieb Exp $

#if !defined(AIPS_ARRAY_H)
#define AIPS_ARRAY_H

#if defined(_AIX)
#pragma implementation ("Array.cc")
#endif

#include <aips/aips.h>
#include <aips/Excp.h>
#include <aips/Block.h>
#include <aips/CountedPtr.h>
#include <aips/ArrayRtti.h>
#include <aips/ArrayIO.h>
#include <aips/Arrays/LogiArray.h>
#include <aips/Arrays/MaskLogiArrFwd.h>

//# Forwards
class AipsIO;
template<class T> class ArrayIterator;
class IPosition;

template<class T> class ROMaskedArray;
template<class T> class MaskedArray;

// 
// <summary> A templated N-D Array class with variable origin </summary>

// Array<T> is a templated, N-dimensional, Array class. The origin is variable,
// but by default indices are zero-based. This Array class is the
// base class for specialized Vector<T>, Matrix<T>, and Cube<T> classes.
//
// Indexing into the array, and positions in general, are given with IPosition
// (essentially a vector of integers) objects. That is, an N-dimensional 
// array requires a length-N IPosition to define a position within the array.
// Unlike C, indexing is done with (), not []. Also, the storage order
// is the same as in FORTRAN, i.e. memory varies most rapidly with the first
// index.
// <srcblock>
//                                     // axisLengths = [1,2,3,4,5]
// IPosition axisLengths(5, 1, 2, 3, 4, 5); 
// Array<Int> ai(axisLengths);         // ai is a 5 dimensional array of
//                                     // integers; indices are 0-based
//                                     // => ai.nelements() == 120
// IPosition origin(5); origin = -1;   //    origin = [-1,-1,-1,-1,-1]
// Array<Int> ai2(axisLengths, origin);// The first element is at index
//                                     // "origin".
// IPosition zero(5); zero = 0;        // [0,0,0,0,0]
// //...
// ai(zero) = ai2(origin);             // Copy first element of ai2 to ai1
// </srcblock>
// Indexing into an N-dimensional array is relatively expensive. Normally
// you will index into a Vector, Matrix, or Cube. These may be obtained from
// an N-dimensional array by creating a reference, or by using an 
// ArrayIterator. The "shape" of the array is an IPosition which gives the
// length of each axis, and the "origin" of the array is an IPosition which
// gives the index of the first array element ([0,0,0...] by default).
//
// An Array may be standalone, or it may refer to another array, or to
// part of another array (by refer we mean that if you change a pixel in 
// the current array, a pixel in the referred to array also changes, i.e.
// they share underlying storage).
// <note role=warning> One way one array can reference another is through the copy 
//        constructor. While this might be what you want, you should
//        probably use the reference() member function to make it explicit.
//        The copy constructor is used when arguments are passed by value;
//        normally functions should not pass Arrays by value, rather they
//        should pass a reference or a const reference. On the positive
//        side, returning an array from a function is efficient since no
//        copying need be done. Later releases of the array classes might
//        have the copy constructor actually make a copy -- comments solicited.
//
// Aside from the explicit reference() member function, a user will
// most commonly encounter an array which references another array
// when he takes an array slice (or section). A slice is a sub-region of
// an array (which might also have a stride: every nth row, every mth column,
// ...).
// <srcblock>
// IPosition lengths(3,10,20,30);
// Array<Int> ai(lengths);         // A 10x20x30 cube
// Cube<Int> ci;
// //...
// ci.reference(ai1);              // ci and ai now reference the same
//                                 // storage
// ci(0,0,0) = 123;                // Can use Cube indexing
// ci.xyPlane(2) = 0;              //     and other member functions
// IPosition zero(3,0,0,0);
// assert(ai(zero) == 123);        // True because ai, ci are references
// //...
// Array<Int> subArray;
// IPosition blc(3,0,0,0), trc(3,5,5,5);
// subArray.reference(ai(blc, trc));
// subArray = 10;                  // All of subArray, which is the
//                                 // subcube from 0,0,0 to 5,5,5 in
//                                 // ai, has the value 10.
// </srcblock>
// While the last example has an array slice referenced explicitly by another
// array variable, normally the user will often only use the slice as
// a temporary in an expresion, for example:
// <srcblock>
//   Array<Complex> array;
//   IPosition blc, trc, offset;
//   //...
//   // Copy from one region of the array into another
//   array(blc, trc) = array(blc+offset, trc+offset);
// </srcblock>
//
// The Array classes are intended to operate on relatively large
// amounts of data. While they haven't been extensively tuned yet,
// they are relatively efficient in terms of speed. Presently they
// are not space efficient -- the overhead is about 15 words. While
// this will be improved (probably to about 1/2 that), these array
// classes are not appropriate for very large numbers of very small
// arrays. The Block<T> class may be what you want in this circumstance.
//
// Element by element mathematical and logical operations are available
// for arrays (defined in aips/ArrayMath.h and aips/ArrayLogical.h).
// Because arithmetic and logical functions are split out, it is possible
// to create an Array<T> (and hence Vector<T> etc) for any type T that has
// a default constructor, assignment operator, and copy constructor. In
// particular, Array<String> works.
//
// Array<T> is derived from Cleanup, so if an exception is thrown storage
// from the array will be reclaimed. When "real" exceptions are available,
// this will no longer be required.
//
// If compiled with the preprocessor symbol AIPS_DEBUG symbol, array
// consistency ("invariants") will be checked in most member
// functions, and indexing will be range-checked. This should not be
// defined for production runs.
//
// A tutorial for the ArrayClasses is available in the "AIPS++ Programming
// Manual."
//
// <note role=tip>
// Most of the data members and functions which are "protected" should
// likely become "private".

template<class T> class Array : public Cleanup
{
public:
    // Result has dimensionality of one, but nelements is zero. Perhaps
    // instead it should be a dimension zero, length zero array.
    Array();

    // Create an array of the given shape, i.e. after construction
    // array.ndim() == shape.nelements() and array.shape() == shape.
    // The origin of the Array is zero.
    Array(const IPosition &shape);

    // Create an array with a given shape and origin.
    Array(const IPosition &shape, const IPosition &origin);

    // After construction, this and other reference the same storage.
    Array(const Array<T> &other);

    // Frees up storage only if this array was the last reference to it.
    virtual ~Array();

    // This functions is used by the exception handling mechanism we have
    // defined. It merely calls the destructor. When real exceptions are
    // available it will be unnecessary.
    //*display 8
    void cleanup();

    // Set every element of the array to "value." Also could use the
    // assignment operator which assigns an array from a scalar.
    void set(const T &value);

    // Apply the function to every element of the array. This modifies
    // the array in place.
    // <group>
    // This version takes a function which takes a T and returns a T.
    void apply(T (*function)(T));
    // This version takes a function which takes a const T reference and
    // returns a T.
    //*display 6
    void apply(T (*function)(const T &));
    // </group>

    // After invocation, this array and other reference the same
    // storage. That is, modifying an element through one will show
    // up in the other. The arrays appear to be identical;
    // they have the same shape and origin.
    virtual void reference(Array<T> &other);

    // Copy the values in other to this. If the array on the left hand
    // side has no elements, then it is resized to be the same size as
    // as the array on the right hand side. Otherwise, the arrays must
    // conform (same shapes, but origins do not need to be the same).
    // <srcblock>
    // IPosition shape(2,10,10);     // some shape
    // Array<Double> ad(shape);
    // //...
    // Array<Double> ad2;            // N.B. ad2.nelements() == 0
    // ad2 = ad;                     // ad2 resizes, then elements
    //                               //     are copied.
    // shape = 20;
    // Array<Double> ad3(shape);
    // ad3 = ad;                     // Error: arrays do not conform
    // </srcblock>
    virtual Array<T> &operator=(const Array<T> &other);

    // Set every element of this array to "value". In other words, a scalar
    // behaves as if it were a constant conformant array.
    Array<T> &operator=(const T &value);

    // Copy to this those values in marray whose corresponding elements
    // in marray's mask are True.
    //
    // <thrown>
    //    <li> ArrayConformanceError
    // </thrown>
    //
    Array<T> &operator= (const ROMaskedArray<T> &marray);

    // This makes a copy of the array and returns it. This can be
    // useful for, e.g. making working copies of function arguments
    // that you can write into.
    // <srcblock>
    // void someFunction(const Array<Int> &arg)
    // {
    //     Array<Int> tmp(arg.copy());
    //     // ...
    // }
    // </srcblock>
    // Note that since the copy constructor makes a reference, if we just
    // created used to copy constructor, modifying "tmp" would also
    // modify "arg". Clearly another alternative would simply be:
    // <srcblock>
    // void someFunction(const Array<Int> &arg)
    // {
    //     Array<Int> tmp;
    //     tmp = arg;
    //     // ...
    // }
    // </srcblock>
    // which likely would be simpler to understand. (Should copy() 
    // be deprecated and removed?)
    Array<T> copy() const;                         // Make a copy of this

    // This ensures that this array does not reference any other storage.
    // <note role=tip> When a section is taken of an array with non-unity strides,
    //        storage can be wasted if the array which originally contained
    //        all the data goes away. unique() also reclaims storage. This
    //        is an optimization users don't normally need to understand.
    //
    //        <srcblock>
    //        IPosition shape(...), blc(...), trc(...), inc(...);
    //        Array<Float> af(shape);
    //        inc = 2; // or anything > 1
    //        Array<Float> aSection.reference(af(blc, trc, inc));
    //        af.reference(anotherArray);
    //        // aSection now references storage that has a stride
    //        // in it, but nothing else is. Storage is wasted.
    //        aSection.unique();
    //        </srcblock>
    void unique();

    // It is occasionally useful to have an array which access the same
    // storage appear to have a different shape and origin. For example,
    // Turning an N-dimensional array into a Vector. This will work when
    // before.nelements() == after.nelements() and if this array refers
    // to other storage, the increment on all axes must be one (guaranteed,
    // for example, by unique()).
    // <srcblock>
    // IPosition squareShape(2,5,5);
    // Array<Float> square(squareShape);
    // IPosition lineShape(1,25);
    // Vector<Float> line(square.reform(lineShape));
    // // "square"'s storage may  now be accessed through Vector "line"
    // </srcblock>
    Array<T> reform(const IPosition &shape, const IPosition &origin) const;
    
    // This member function returns an Array reference which has all degenerate
    // (length==1) axes removed and the dimensionality reduced appropriately.
    // If nelements==0 returns a dimension 1 0-length array.
    // Otherwise returns an array with all length-1 axes remoeved.
    Array<T> nonDegenerate();

    // Make this array a different shape. Presently the old values are not
    // copied over to the new array.
    // <group>
    virtual void resize(const IPosition &newShape);
    virtual void resize(const IPosition &newShape, const IPosition &newOrigin);
    // </group>

    // Access a single element of the array. This is relatively
    // expensive. Extensive indexing should be done through one
    // of the Array specializations (Vector, Matrix, Cube). If
    // AIPS_DEBUG is defined, index checking will be performed.
    // <group>
    T &operator()(const IPosition &);
    //*display 6
    const T &operator()(const IPosition &) const;
    // </group>

    // Ret a reference to an array which extends from "start" to end."
    //
    // <group>

    Array<T> operator()(const IPosition &start, const IPosition &end);

    // Along the ith axis, every inc[i]'th element is chosen.
    Array<T> operator()(const IPosition &start, const IPosition &end,
			const IPosition &inc);
    // </group>


    // The array is masked by the input LogicalArray.
    // This mask must conform to the array, but it does not need to have the
    // same origin.
    //
    // <group>

    // Return a ROMaskedArray.
    ROMaskedArray<T> operator() (const LogicalArray &mask) const;

    // Return a ROMaskedArray.
    MaskedArray<T> operator() (const LogicalArray &mask);
 
    // </group>

    
    // The array is masked by the input ROMaskedLogicalArray.
    // The mask is effectively the AND of the internal LogicalArray
    // and the internal mask of the ROMaskedLogicalArray.
    // The ROMaskedLogicalArray must conform to the array, but it does not
    // need to have the same origin.
    //
    // <group>

    // Return a ROMaskedArray.
    ROMaskedArray<T> operator() (const ROMaskedLogicalArray &mask) const;

    // Return a ROMaskedArray.
    MaskedArray<T> operator() (const ROMaskedLogicalArray &mask);
 
    // </group>

    
    // The number of references the underlying storage has assigned to it.
    // It is 1 unless there are outstanding references to the storage (e.g.,
    // through a slice). Normally you have no need to do this since the
    // arrays handle all of the references for you.
    uInt nrefs() const;

    // The dimensionality of this array.
    uInt ndim() const {return ndimen;}

    // How many elements does this array have? Product of all axis lengths.
    uInt nelements() const { return nels; }

    // Check to see if the Array is consistent. This is about the same thing
    // as checking for invariants. If AIPS_DEBUG is defined, this is invoked
    // after construction and on entry to most member functions.
    virtual Bool ok() const;

    // Are the shapes identical? The origins do NOT need to be the same.
    // Binary operations will "line up" the arrays at their origins, so
    // as long as the shapes are the same the Arrays conform.
    // <group>
    Bool conform(const Array<T> &other) const;
    Bool conform(const ROMaskedArray<T> &other) const;
    // </group>

    // The IPosition of the first array element. All zero is the default.
    // When zero based, max_index(i) == shape(i) - 1.
    IPosition origin() const;
    // The length of each axis.
    IPosition shape() const;
    // A convenience function: end(i) = origin(i) + shape(i) - 1; i.e. this
    // is the IPosition of the last element of the Array.
    IPosition end() const;

    // Generally use of this should be shunned, except to use a FORTRAN routine
    // or something similar. Because you can't know the state of the underlying
    // data layout (in particular, if there are increments) sometimes the
    // pointer returned will be to a copy, but often this won't be necessary.
    // A boolean is returned which tells you if this is a copy (and hence the
    // storage must be deleted). Note that if you don't do anything unusual,
    // getStorage followed by putStorage will do the deletion for you (if
    // required). e.g.:
    // <srcblock>
    // Array<Int> a(shape); ...
    // Bool deleteIt; Int *storage = a.getStorage(deleteIt);
    // foo(storage, a.nelements()); a.puStorage(storage, deleteIt);
    // // or a.freeStorage(storage, deleteIt) if a is const.
    // </srcblock>
    // NB: However, if you only use getStorage, you will have to delete the
    // pointer yourself using freeStorage().
    //
    // It would probably be useful to have corresponding "copyin" "copyout"
    // functions that used a user supplied buffer.
    // Note that deleteIt is set in this function.
    // <group>
    T *getStorage(Bool &deleteIt);
    //*display 6
    const T *getStorage(Bool &deleteIt) const;
    // </group>

    // putStorage() is normally called after a call to getStorage() (cf).
    // The "storage" pointer is set to zero.
    void putStorage(T *&storage, Bool deleteAndCopy);

    // If deleteIt is set, delete "storage". Normally freeStorage calls
    // will follow calls to getStorage. The reason the pointer is "const"
    // is because only const pointers are released from const arrays.
    // The "storage" pointer is set to zero.
    void freeStorage(const T *&storage, Bool deleteIt) const;

    // Used to iterate through Arrays. Derived classes VectorIterator and
    // MatrixIterator are probably more useful.
    friend class ArrayIterator<T>;

    // Array version for major change (used by ArrayIO).
    // enum did not work properly with cfront 3.0.1), so replaced
    // by a static inline function. Users won't normally use this.
    //*display 8
    static uInt arrayVersion() {return 2;}

    // Macro to define the typeinfo member functions.
    //*display 8
    rtti_dcl_mbrf_p1(Array<T>, Cleanup);

protected:
    // Number of elements in the array. Cached rather than computed.
    uInt nels;

    // Dimensionality of the array.
    int ndimen;

    // Used to hole the origin, shape, increment into the underlying storage
    // and originalLength of the array. These should be coelesced to avoid
    // wasted space by we have by using 4 blocks.
    Block<Int> start, length, inc, originalLength;

    // Reference counted block that contains the storage.
    CountedPtr<Block<T> > data;

    // This pointer is adjusted to point to the first element of the array.
    // It is not necessarily the same thing as data->storage() since
    // this array might be a section, e.g. have a blc which shifts us forward
    // into the block.
    T *begin;

    // Various helper functions that should be deleted.
    //*display 8
    // <group>
    Array(uInt, const Block<Int> &);
    void validateConformance(const Array<T> &) const;
    void validateIndex(const IPosition &) const;
    // </group>

    // A helper function to copy one vector to another, both of which might
    // have strides. This is probably useful enough that it should be made
    // available more publicly?
    void copyVec(uInt n, T *to, uInt toStride, const T *from, 
		 uInt fromStride) const;

    // Copy a scalar value into a carray with a given stride.
    void Array<T>::copyScalar(uInt n, T *to, uInt toStride, T value) const;
};


// 
// What is the volume of an N-dimensional array. 
// Shape[0]*Shape[1]*...*Shape[N-1]. An Array helper function.
//*display 6
uInt ArrayVolume(uInt Ndim, const Int *Shape);

// 
// What is the linear index into an "Ndim" dimensional array of the given
// "Shape", "Origin", and "Increment" for a given IPosition Index.
//  An Array helper function.
//*display 6
uInt ArrayIndexOffset(uInt Ndim, const Int *Shape, 
		      const Int *Origin, const Int *Inc, 
		      const IPosition &Index);


// 
// Test conformance for two arrays of different types.
// Are the shapes identical? The origins do NOT need to be the same.
// Binary operations will "line up" the arrays at their origins, so
// as long as the shapes are the same the arrays conform.
//
template<class T, class U>
  Bool conform2 (const Array<T> &left, const Array<U> &right);

#endif
