//# Block.h: Simple templated array classes
//# Copyright (C) 1993,1994
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: aips2-request@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id: Block.h,v 1.1.1.1 1996/04/04 16:37:04 dschieb Exp $

#if !defined (AIPS_BLOCK_H)
#define AIPS_BLOCK_H

#if defined(_AIX)
#pragma implementation ("../implement/Block.cc")
#endif

#include <aips/aips.h>
#include <aips/BlockIO.h>

//# For index checking
#if defined(AIPS_ARRAY_INDEX_CHECK)
#include <aips/Error.h>
#endif

// 
//
// Block<T> is a simple templated 1-D array class. Indices are always
// 0-based. For efficiency reasons, no index checking is done unless the
// preprocessor symbol AIPS_ARRAY_INDEX_CHECK is defined. Block<T>'s may be
// assigned to and constructed from other Block<T>'s. As no reference counting
// is done this can be an expensive operation, however.
//
// The net effect of this class is meant to be unsurprising to users who think
// of arrays as first class objects. The name "Block" is  intended to convey
// the concept of a solid "chunk" of things without any intervening "fancy"
// memory management, etc. This class was written to be
// used in the implementations of more functional Vector, Matrix, etc. classes,
// although it is expected Block<T> will be useful on its own.
//
//
// The Block class should be efficient. You should normally use Block
// in preference to C-style arrays, particularly those obtained from "new".
// Block<T> is not derived from Cleanup
// so if an exception is thrown the storage won't be reclaimed: Block's should
// be used to build other classes which are themselves reclaimed.
//
// <note role=warning> If you use the assignment operator on an element of this class, you
// will may leave dangling references to pointers released from storage().
// Resizing the array will also have this effect if the underlying storage
// is actually affected.
//
// Currently the AllocError is thrown are in response to new[] failing.
// If index checking is turned on, an out-of-bounds index will
// generate an indexError<uInt> exception.
//
// <h3>Examples</h3> <srcblock>
// Block<Int> a(100,0);  // 100 ints initialized to 0
// Block<Int> b;         // 0-length Block
// // ...
// b = a;                // resize b and copy a into it
// for (uInt i=0; i < a.nelements(); i++) {
//     a[i] = i;    // Generate a sequence
//                  // with Vectors, could simply say "indgen(myVector);"
// }
// b.set(-1);       // All positions in b have the value -1
// b.resize(b.nelements()*2); // Make b twice as long, by default the old
//                            // elements are copied over, although this can
//                            // be defeated.
// some_c_function(b.storage());  // Use a fn that takes an Int * pointer
// </srcblock>

// <descriptor>
// <execution> <sequential>
// </descriptor>

template<class T> class Block {
public:

    // Create a zero-length Block. Note that any index into this Block
    // is an error.
    Block();
    // Create a Block with the given number of points. The values in Block
    // are uninitialized. Note that indices range between 0 and n-1.
    Block(uInt n);
    // Create a Block of the given length, and initialize (via operator= for 
    // objects of type T) with the provided value.
    Block(uInt n, T val);
    // Create a Block from a C-array (i.e. pointer). If takeOverStorage is
    // True, The Block assumes that it
    // owns the pointer, i.e. that it is safe to delete[] it when the Block
    // is destructed, otherwise the actual storage is not destroyed. If true,
    // storagePointer is set to NULL.
    Block(uInt n, T *&storagePointer, Bool takeOverStorage = True);
    // Copy the other block into this one. Uses copy, not reference, semantics.
    Block(const Block<T> &other);
  
    // Assign other to this. this resizes itself to the size of other, so after
    // the assignment, this->nelements() == other.elements() always.
    Block<T> &operator=(const Block<T> &other);

    // Frees up the storage pointed contained in the Block.
    ~Block();

    // Resizes the Block. If "n == nelements()" resize just returns. If
    // A larger size is requested ("n > nelements()") the Block always
    // resizes. If the requested size is smaller ("n < nelements()"),
    // by default the Block does not resize smaller, although it can be
    // forced to with "forceSmaller". The reasoning behind this is that
    // often the user will just want a buffer of at least a certain size,
    // and won't want to pay the cost of multiple resizings.
    // <srcblock>
    // Block<float> bf(100, 0.0);
    // bf.resize(10);        // bf.nelements() == 100
    // bf.resize(10, True)   // bf.nelements() == 10
    // bf.resize(200)        // bf.nelements() == 200
    // </srcblock>
    // Normally the old elements are copied over (although if the
    // Block is lengthened the trailing elements will have undefined
    // values), however this can be turned off by setting copyElements to
    // False.
    //
    // This is written as three functions because default parameters do
    // not always work properly with templates.
    // <group>
    void resize(uInt n, Bool forceSmaller, Bool copyElements);
    void resize(uInt n);
    void resize(uInt n, Bool forceSmaller);
    // </group>

    // Remove a single element from the Block. This forces a new allocations
    // and copy, so it is relatively expensive.
    void remove(uInt whichOne);


    // Replace the internal storage with a C-array (i.e. pointer).
    // If takeOverStorage is True, The Block assumes that it
    // owns the pointer, i.e. that it is safe to delete[] it when the Block
    // is destructed, otherwise the actual storage is not destroyed. If true,
    // storagePointer is set to NULL.
    void replaceStorage(uInt n, T *&storagePointer, Bool takeOverStorage=True);

    // Index into the block (0-based). If the preprocessor symbol
    // AIPS_ARRAY_INDEX_CHECK is defined, index checking will be done
    //and an out-of-bounds index will cause an indexError<uInt> to be
    // thrown. Note that valid indices range between 0 and nelements()-1.
    // <thrown>
    //    <li> indexError
    // </thrown>
    // <group>
    T &operator[](uInt index)
    {
#if defined(AIPS_ARRAY_INDEX_CHECK)
        // Write it this way to avoid casts; remember index and npts are
        // unsigned.
        if ((npts == 0) || (index > npts - 1)) {
	    throw(indexError<uInt>(index, "Block::operator[] - "
				   "index out of range"));
	}
#endif
	return array[index];
    }
    const T &operator[](uInt index) const
    {
#if defined(AIPS_ARRAY_INDEX_CHECK)
        if ((npts == 0) || (index > npts - 1)) {
	    throw(indexError<uInt>(index, "Block::operator[] const - "
				   "index out of range"));
	}
#endif
	return array[index];
    }
    // </group>

    // Set all values in the block to "val".
    void set(const T &val);

    // If you really, really, need a "raw" pointer to the beginning of the
    // storage area this will give it to you. This may leave dangling pointers
    // if the block is destructed or if the assignment operator or resize 
    // is used. Returns a null pointer if nelements() == 0.
    // It is best to only use this if you completely control the extent and
    // lifetime of the Block.
    // <h3> Examples of misuse </h3> <srcblock>
    // Block<int> *bp = new Block<int>(100);
    // int *ip = bp->storage();
    // delete bp;      // Oops, ip is now dangling
    // Block<int> a(100),b(100);
    // int *ip = a.storage();
    // a = b;          // Likewise
    // </srcblock>
    // <group>
    T *storage() {return array;}
    const T *storage() const {return array;}
    // </group>

    // The number of elements contained in this Block<T>.
    uInt nelements() const {return npts;}
private:
    // The number of points in the vector
    uInt  npts;
    // The actual storage
    T *array;
    // Can we delete the storage upon destruction?
    Bool destroyPointer;
};

#endif
