//# MaskArrMath.h: Simple mathematics done with MaskedArray's.
//# Copyright (C) 1993,1994
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: aips2-request@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id: MaskArrMath.h,v 1.1.1.1 1996/04/04 16:37:05 dschieb Exp $

#if !defined(AIPS_MASK_ARR_MATH_H)
#define AIPS_MASK_ARR_MATH_H

#if defined(_AIX)
#pragma implementation ("MaskArrMath.cc")
#endif


#include <aips/aips.h>
#include <aips/Array.h>
#include <aips/Arrays/MaskedArray.h>
#include <aips/IPosition.h>

// Needed to get the proper Complex typedef's
#include <aips/Complex.h>






// 
// Dummy function to simulate section introduction documentation.
//
// <summary>
// Functions for element by element <here>mathematical operations</here> on masked arrays.
// </summary>
// <reviewed reviewer="" date="" tests="">
//
// <prerequisite>
//   <li> Array
//   <li> MaskedArray
// </prerequisite>
//
// <etymology>
// MaskArrMath is short for MaskedArrayMath, which is too long by
// AIPS++ file naming conventions.  This file contains global functions
// which perform element by element mathematical operations on masked arrays.
// </etymology>
//
// <synopsis>
// <a name="info">These functions perform element by element mathematical</a>
// operations on masked arrays.  With two arrays, they must both conform, and
// the result is done element by element, for those locations where the mask 
// of the MaskedArray is True.  For two MaskedArrays, the "and" of the masks
// is used.
// </synopsis>
//
// <example>
// <srcblock>
//   Vector<Int> a(10);
//   Vector<Int> b(10);
//   Vector<Int> c(10);
//      . . .
//   c = a(a>0) + b(b>0);
// </srcblock>
// This example sets those elements of c where ((a>0) && (b>0)) to (a+b).
// Elements of c where !((a>0) && (b>0)) are unchanged.  The result of
// this operation is a MaskedArray.  The assignment from this
// MaskedArray to the Vector c only assigns those elements
// where the mask is True.
// </example>
//
// <example>
// <srcblock>
//   Vector<Double> a(10);
//   Vector<Double> b(10);
//   Vector<Double> c(10);
//      . . .
//   c = atan2 (a, b(b>0);
// </srcblock>
// This example sets those elements of c where (b>0) to atan2 (a,b).
// Elements of c where !(b>0) are unchanged.  The result of
// this operation is a MaskedArray.  The assignment from this
// MaskedArray to the Vector c only assigns those elements
// where the mask is True.
// </example>
//
// <example>
// <srcblock>
//   Vector<Int> a(10);
//   Int result;
//      . . .
//   result = sum (a(a>0));
// </srcblock>
// This example sums a, for those elements of a which are greater than 0.
// </example>
//
// <motivation>
// One wants to be able to mask arrays and perform mathematical operations on
// those masked arrays.  Since the masked arrays are only defined where
// the masks are True, the result must be a MaskedArray, or a simple number.
// </motivation>
//
// <linkfrom anchor=math classes="MaskedArray ROMaskedArray">
//     Related <here>mathematical functions</here> for MaskedArrays
// </linkfrom>
//
// <group name=math>

// 
// Element by element arithmetic modifying left in-place. left and other
// must be conformant.
//
// <thrown>
//   <li> ArrayConformanceError
// </thrown>
//
// <group>
template<class T>
  const MaskedArray<T> & operator+= (const MaskedArray<T> &left,
                                     const Array<T> &other);
template<class T>
  const MaskedArray<T> & operator-= (const MaskedArray<T> &left,
                                     const Array<T> &other);
template<class T>
  const MaskedArray<T> & operator*= (const MaskedArray<T> &left,
                                     const Array<T> &other);
template<class T>
  const MaskedArray<T> & operator/= (const MaskedArray<T> &left,
                                     const Array<T> &other);
//
template<class T>
  Array<T> & operator+= (Array<T> &left, const ROMaskedArray<T> &other);
template<class T>
  Array<T> & operator-= (Array<T> &left, const ROMaskedArray<T> &other);
template<class T>
  Array<T> & operator*= (Array<T> &left, const ROMaskedArray<T> &other);
template<class T>
  Array<T> & operator/= (Array<T> &left, const ROMaskedArray<T> &other);
//
template<class T>
  const MaskedArray<T> & operator+= (const MaskedArray<T> &left,
                                     const ROMaskedArray<T> &other);
template<class T>
  const MaskedArray<T> & operator-= (const MaskedArray<T> &left,
                                     const ROMaskedArray<T> &other);
template<class T>
  const MaskedArray<T> & operator*= (const MaskedArray<T> &left,
                                     const ROMaskedArray<T> &other);
template<class T>
  const MaskedArray<T> & operator/= (const MaskedArray<T> &left,
                                     const ROMaskedArray<T> &other);
// </group>

// 
// Element by element arithmetic modifying left in-place. The scalar "other"
// behaves as if it were a conformant Array to left filled with constant values.
// <group>
template<class T>
  const MaskedArray<T> & operator+= (const MaskedArray<T> &left,
                                     const T &other);
template<class T>
  const MaskedArray<T> & operator-= (const MaskedArray<T> &left,
                                     const T &other);
template<class T>
  const MaskedArray<T> & operator*= (const MaskedArray<T> &left,
                                     const T &other);
template<class T>
  const MaskedArray<T> & operator/= (const MaskedArray<T> &left,
                                     const T &other);
// </group>

// Unary arithmetic operation.
// 
// <group>
template<class T> MaskedArray<T> operator+(const ROMaskedArray<T> &a);
template<class T> MaskedArray<T> operator-(const ROMaskedArray<T> &a);
// </group>

// 
// Element by element arithmetic on MaskedArrays, returning a MaskedArray.
//
// <thrown>
//   <li> ArrayConformanceError
// </thrown>
//
// <group>
template<class T> 
  MaskedArray<T> operator+ (const ROMaskedArray<T> &left, const Array<T> &right);
template<class T> 
  MaskedArray<T> operator- (const ROMaskedArray<T> &left, const Array<T> &right);
template<class T> 
  MaskedArray<T> operator* (const ROMaskedArray<T> &left, const Array<T> &right);
template<class T> 
  MaskedArray<T> operator/ (const ROMaskedArray<T> &left, const Array<T> &right);
//
template<class T> 
  MaskedArray<T> operator+ (const Array<T> &left, const ROMaskedArray<T> &right);
template<class T> 
  MaskedArray<T> operator- (const Array<T> &left, const ROMaskedArray<T> &right);
template<class T> 
  MaskedArray<T> operator* (const Array<T> &left, const ROMaskedArray<T> &right);
template<class T> 
  MaskedArray<T> operator/ (const Array<T> &left, const ROMaskedArray<T> &right);
//
template<class T> 
  MaskedArray<T> operator+ (const ROMaskedArray<T> &left,
                            const ROMaskedArray<T> &right);
template<class T> 
  MaskedArray<T> operator- (const ROMaskedArray<T> &left,
                            const ROMaskedArray<T> &right);
template<class T> 
  MaskedArray<T> operator* (const ROMaskedArray<T> &left,
                            const ROMaskedArray<T> &right);
template<class T> 
  MaskedArray<T> operator/ (const ROMaskedArray<T> &left,
                            const ROMaskedArray<T> &right);
// </group>

// 
// Element by element arithmetic between a MaskedArray and a scalar, returning
// a MaskedArray.
// <group>
template<class T> 
    MaskedArray<T> operator+ (const ROMaskedArray<T> &left, const T &right);
template<class T> 
    MaskedArray<T> operator- (const ROMaskedArray<T> &left, const T &right);
template<class T> 
    MaskedArray<T> operator* (const ROMaskedArray<T> &left, const T &right);
template<class T> 
    MaskedArray<T> operator/ (const ROMaskedArray<T> &left, const T &right);
// </group>

// 
// Element by element arithmetic between a scalar and a MaskedArray, returning
// a MaskedArray.
// <group>
template<class T>  
    MaskedArray<T> operator+ (const T &left, const ROMaskedArray<T> &right);
template<class T>  
    MaskedArray<T> operator- (const T &left, const ROMaskedArray<T> &right);
template<class T>  
    MaskedArray<T> operator* (const T &left, const ROMaskedArray<T> &right);
template<class T>  
    MaskedArray<T> operator/ (const T &left, const ROMaskedArray<T> &right);
// </group>

// 
// Transcendental function applied to the array on an element-by-element
// basis. Although a template function, this may not make sense for all
// numeric types.
// <group name=transcendental>
template<class T> MaskedArray<T> sin(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> cos(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> tan(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> asin(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> acos(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> atan(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> sinh(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> cosh(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> tanh(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> exp(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> log(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> log10(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> sqrt(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> abs(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> fabs(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> ceil(const ROMaskedArray<T> &left);
template<class T> MaskedArray<T> floor(const ROMaskedArray<T> &left);

// <thrown>
//   <li> ArrayConformanceError
// </thrown>
//
// <group>

template<class T>
  MaskedArray<T> atan2(const ROMaskedArray<T> &left, const Array<T> &right);
template<class T>
  MaskedArray<T> fmod(const ROMaskedArray<T> &left, const Array<T> &right);

template<class T>
  MaskedArray<T> atan2(const Array<T> &left, const ROMaskedArray<T> &right);
template<class T>
  MaskedArray<T> fmod(const Array<T> &left, const ROMaskedArray<T> &right);

template<class T>
  MaskedArray<T> atan2(const ROMaskedArray<T> &left,
                       const ROMaskedArray<T> &right);
template<class T>
  MaskedArray<T> fmod(const ROMaskedArray<T> &left,
                      const ROMaskedArray<T> &right);

// </group>

template<class T>
  MaskedArray<T> atan2(const ROMaskedArray<T> &left, const T &right);
template<class T>
  MaskedArray<T> fmod(const ROMaskedArray<T> &left, const T &right);

template<class T>
  MaskedArray<T> atan2(const T &left, const ROMaskedArray<T> &right);
template<class T>
  MaskedArray<T> fmod(const T &left, const ROMaskedArray<T> &right);

// <thrown>
//   <li> ArrayConformanceError
// </thrown>
//
// <group>

template<class T, class U>
  MaskedArray<T> pow(const ROMaskedArray<T> &left, const Array<U> &right);
template<class T, class U>
  MaskedArray<T> pow(const Array<T> &left, const ROMaskedArray<U> &right);
template<class T, class U>
  MaskedArray<T> pow(const ROMaskedArray<T> &left,
                     const ROMaskedArray<U> &right);

// </group>

template<class T>
  MaskedArray<T> pow(const ROMaskedArray<T> &left, const Double right);

// </group>


// 
// Find the minimum and maximum values of a MaskedArray.
// <group>

// Also find the IPositions of the minimum and maximum values.
//
// <thrown>
//    <li> ArrayError
// </thrown>
//
template<class T>
void minMax(T &minVal, T &maxVal, IPosition &minPos, IPosition &maxPos,
            const ROMaskedArray<T> &marray);

template<class T>
void minMax(T &minVal, T &maxVal,
            const ROMaskedArray<T> &marray);

// </group>


// 
// The "min" and "max" functions require that the type "T" have comparison 
// operators.
// <group>

// The minimum element of the array.
template<class T> T min(const ROMaskedArray<T> &left);


// Return an array that contains the minimum of "left" and "right" at each
// position.
//
// <group>

// "left" and "right" must be conformant.
//
// <thrown>
//    <li> ArrayError
// </thrown>
//
// <group>

template<class T>
MaskedArray<T> min(const ROMaskedArray<T> &left, const Array<T> &right);

template<class T>
MaskedArray<T> min(const Array<T> &left, const ROMaskedArray<T> &right);

template<class T>
MaskedArray<T> min(const ROMaskedArray<T> &left, const ROMaskedArray<T> &right);

// </group>

template<class T>
MaskedArray<T> min(const T &left, const ROMaskedArray<T> &right);

template<class T>
MaskedArray<T> min(const ROMaskedArray<T> &left, const T &right);

// </group>


// "result" contains the minimum of "left" and "right" at each position.
// "result", "left", and "right" must be conformant.
//
// <thrown>
//   <li> ArrayConformanceError
// </thrown>
//
template<class T> void min(const MaskedArray<T> &result,
                           const Array<T> &left, const Array<T> &right);


// The maximum element of the array.
template<class T> T max(const ROMaskedArray<T> &left);


// Return an array that contains the maximum of "left" and "right" at each
// position.
//
// <group>

// "left" and "right" must be conformant.
//
// <thrown>
//    <li> ArrayError
// </thrown>
//
// <group>

template<class T>
MaskedArray<T> max(const ROMaskedArray<T> &left, const Array<T> &right);

template<class T>
MaskedArray<T> max(const Array<T> &left, const ROMaskedArray<T> &right);

template<class T>
MaskedArray<T> max(const ROMaskedArray<T> &left, const ROMaskedArray<T> &right);

// </group>

template<class T>
MaskedArray<T> max(const T &left, const ROMaskedArray<T> &right);

template<class T>
MaskedArray<T> max(const ROMaskedArray<T> &left, const T &right);

// </group>


// "result" contains the maximum of "left" and "right" at each position.
// "result", "left", and "right" must be conformant.
//
// <thrown>
//   <li> ArrayConformanceError
// </thrown>
//
template<class T> void max(const MaskedArray<T> &result,
                           const Array<T> &left, const Array<T> &right);

// </group>



// 
// Fills all elements of "array" where the mask is True with a sequence
// starting with "start" and incrementing by "inc" for each element
// where the mask is True.
// The first axis varies most rapidly.
template<class T> void indgen(const MaskedArray<T> &a,
                              const T start, const T inc);

// 
// Fills all elements of "array" where the mask is True with a sequence
// starting with 0 and incremented by one for each element
// where the  mask is True.
// The first axis varies most rapidly.
template<class T>  void indgen(const MaskedArray<T> &a);

// 
// Fills all elements of "array" where the mask is True with a sequence
// starting with "start" and incremented by one for each element
// where the  mask is True.
// The first axis varies most rapidly.
template<class T>  void indgen(const MaskedArray<T> &a, const T start);


// <thrown>
//    <li> ArrayError
// </thrown>
//
// <group>

// 
// Sum of every element of the MaskedArray where the Mask is True.
template<class T> T sum(const ROMaskedArray<T> &a);

// 
// Sum of the squares of every element of the MaskedArray where the Mask is True.
template<class T> T sumsquares(const ROMaskedArray<T> &a);

// 
// Product of every element of the MaskedArray where the Mask is True.
// This could of course easily overflow.
template<class T> T product(const ROMaskedArray<T> &a);

// 
// The mean of "a" is the sum of all elements of "a" divided by the number
// of elements of "a".
template<class T> T mean(const ROMaskedArray<T> &a);

// 
// The variance of "a" is the sum of (a(i) - mean(a))**2/(a.nelements() - 1).
// N.B. N-1, not N in the denominator).
template<class T> T variance(const ROMaskedArray<T> &a);

// 
// The variance of "a" is the sum of (a(i) - mean(a))**2/(a.nelements() - 1).
// N.B. N-1, not N in the denominator).
// Rather than using a computed mean, use the supplied value.
template<class T> T variance(const ROMaskedArray<T> &a, T mean);

// 
// The standard deviation of "a" is the sqare root of its variance.
template<class T> T stddev(const ROMaskedArray<T> &a);

// 
// The standard deviation of "a" is the sqare root of its variance.
// Rather than using a computed mean, use the supplied value.
template<class T> T stddev(const ROMaskedArray<T> &a, T mean);

// 
// The average deviation of "a" is the sum of abs(a(i) - mean(a))/N. (N.B.
// N, not N-1 in the denominator).
template<class T> T avdev(const ROMaskedArray<T> &a);

// 
// The average deviation of "a" is the sum of abs(a(i) - mean(a))/N. (N.B.
// N, not N-1 in the denominator).
// Rather than using a computed mean, use the supplied value.
template<class T> T avdev(const ROMaskedArray<T> &a,T mean);

// 
// The median of "a" is a((n+1/2)) if a has an odd number of elements,
// otherwise it is 1/2(a(n/2) + a((n+1)/2)).
// If "sorted"==True we assume
// the data is already sorted and we compute the median directly, otherwise
// we sort it first (there are algorithms which don't require a sort we
// could implement if necessary).
template<class T> T median(const ROMaskedArray<T> &a, Bool sorted);

// 
// The median of "a" is a((n+1/2)) if a has an odd number of elements,
// otherwise it is 1/2(a(n/2) + a((n+1)/2)). Assumes that "a" is unsorted.
template<class T> T median(const ROMaskedArray<T> &a);

// </group>


// Returns a MaskedArray where every element is multiplied by itself.
template<class T> MaskedArray<T> square(const ROMaskedArray<T> &val);

// Returns a MaskedArray where every element is multiplied by itself.
template<class T> MaskedArray<T> cube(const ROMaskedArray<T> &val);

// </group>

#endif
