//# MaskedArray.h: A templated N-D masked array class with variable origin.
//# Copyright (C) 1993,1994
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: aips2-request@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id: MaskedArray.h,v 1.1.1.1 1996/04/04 16:37:04 dschieb Exp $

#if !defined (AIPS_MASKEDARRAY_H)
#define AIPS_MASKEDARRAY_H

#if defined(_AIX)
#pragma implementation ("MaskedArray.cc")
#endif



//# Includes
#include <aips/aips.h>
#include <aips/Excp.h>
#include <aips/IPosition.h>
#include <aips/Complex.h>
#include <aips/String.h>
#include <aips/Arrays/LogiArrayFwd.h>
#include <aips/Arrays/MaskLogiArrFwd.h>

//# Forward declarations
//# NOTE:  Array.h cannot be included in this header file.  Anything needed
//#        from it must be forwarded.  This is why LogicalArrayFwd.h is
//#        included instead of LogicalArray.h .
template <class T> class Array;
template <class T> class ROMaskedArray;
template <class T> class MaskedArray;


// 
// <summary>
// Class for masking an Array for readonly operations on that Array.
// </summary>
// <reviewed reviewer="" date="" tests="">
//
// <prerequisite>
//   <li> Array
//   <li> LogicalArray
// </prerequisite>
//
// <etymology>
// ROMaskedArray is a readonly MaskedArray.  It is a MaskedArray where the
// underlying Array is not allowed to be modified by any operations on
// the ROMaskedArray.  See the discussion of MaskedArray for more information
// on MaskedArrays.
// </etymology>
//
// <synopsis>
// A ROMaskedArray is a masked array where the underlying Array is not allowed
// to be modified by any operations on the ROMaskedArray.  The MaskedArray
// class, which does allow the underlying Array to be modified, is derived
// from this class.  See the discussion of MaskedArray for more information
// on MaskedArrays.
//
// Many mathematical and logical global operators and functions which operate
// on ROMaskedArrays are defined. See the discussion in MaskedArrays for
// lists or click <linkto group="MaskArrMath.h#math">here</linkto> 
// (<linkto group="MaskArrMath.h#math:description"></em>description</em></linkto>) to go
// to the functions. (<linkto group="MaskArrMath.h#math#info"><em>info</em></linkto>)
// (<em><linkto file="MaskArrMath.h#info">info</linkto> again!</em>)
// </synopsis>
//
// <example>
// <srcblock>
//   template<class T>
//     const MaskedArray<T> & operator+= (const MaskedArray<T> &left,
//                                        const ROMaskedArray<T> &other);
// </srcblock>
// The operator operator+=() adds the second argument to the first argument,
// putting the result in the first argument.  Therefore, the second argument
// can be a ROMaskedArray, since it is not modified, while the first argument
// must be a MaskedArray, since it is modified.
// </example>
//
// <motivation>
// A MaskedArray is an association between an Array and a LogicalArray which
// masks the Array.  It was necessary that the meaning of const MaskedArray
// be that the association was const, but that the Array itself could be
// modified inside the MaskedArray.  Therefore, it was necessary that there
// be a separate class for masked arrays where the Array itself could not
// be modified.
// </motivation>
//
// <todo asof="$DATE:$>
//   <li> Consider whether the data items should be private instead of
//          protected.
//   <li> Consider whether there should be constructors for masks
//          specified as Array<Bool>.
// </todo>


template<class T> class ROMaskedArray : public Cleanup
{

public:

    // Create a ROMaskedArray from an Array and a LogicalArray.
    //
    // The internal mask is a total copy of the input mask, and is
    // completely independent of the input mask.  However, the origin of
    // the internal mask is shifted to coincide with that of the Array.
    //
    // The Array is copy constructed, which means that it is a really smart
    // pointer to the underlying Block, and shares this Block with the input
    // Array.
    //
    // <thrown>
    //    <li> ArrayConformanceError
    //    <li> AllocError
    // </thrown>
    //
    ROMaskedArray(const Array<T> &inarray, const LogicalArray &inmask);

    // Create a ROMaskedArray from a ROMaskedArray and a LogicalArray.
    //
    // The internal mask is the AND of the input mask and the mask of
    // the input ROMaskedArray.  The origin of the internal mask is set
    // to coincide with that of the Array of the input ROMaskedArray.
    //
    // The Array from the input ROMaskedArray is copy constructed, which
    // means that it is a really smart pointer to the underlying Block, and
    // shares this Block with the Array from the input ROMaskedArray.
    //
    //
    // <thrown>
    //    <li> ArrayConformanceError
    //    <li> AllocError
    // </thrown>
    //
    ROMaskedArray(const ROMaskedArray<T> &inarray, const LogicalArray &inmask);

    // Create a ROMaskedArray from an Array and a ROMaskedLogicalArray.
    //
    // The internal mask is the AND of the internal LogicalArray and the
    // internal mask of the ROMaskedLogicalArray.  The origin of the internal
    // mask is set to coincide with that of the Array of the input
    // ROMaskedArray.
    //
    // The Array is copy constructed, which means that it is a really smart
    // pointer to the underlying Block, and shares this Block with the input
    // Array.
    //
    // <thrown>
    //    <li> ArrayConformanceError
    //    <li> AllocError
    // </thrown>
    //
    ROMaskedArray(const Array<T> &inarray, const ROMaskedLogicalArray &inmask);

    // Create a ROMaskedArray from a ROMaskedArray and a ROMaskedLogicalArray.
    //
    // The internal mask is the AND of the internal LogicalArray and the
    // internal mask of the ROMaskedLogicalArray, ANDed with the mask of
    // the input ROMaskedArray.  The origin of the internal mask is set
    // to coincide with that of the Array of the input ROMaskedArray.
    //
    // The Array from the input ROMaskedArray is copy constructed, which
    // means that it is a really smart pointer to the underlying Block, and
    // shares this Block with the Array from the input ROMaskedArray.
    //
    //
    // <thrown>
    //    <li> ArrayConformanceError
    //    <li> AllocError
    // </thrown>
    //
    ROMaskedArray(const ROMaskedArray<T> &inarray,
                  const ROMaskedLogicalArray &inmask);

    // Copy constructor.
    //
    // The internal mask is a total copy of the mask from the input
    // ROMaskedArray, and is completely independent of this input mask.
    //
    // The Array from the input ROMaskedArray is copy constructed, which
    // means that it is a really smart pointer to the underlying Block, and
    // shares this Block with the Array from the input ROMaskedArray.
    //
    // <thrown>
    //    <li> AllocError
    // </thrown>
    //
    ROMaskedArray(const ROMaskedArray<T> &other);

    ~ROMaskedArray();

    // This functions is used by the exception handling mechanism we have
    // defined. It merely calls the destructor. When real exceptions are
    // available it will be unnecessary.
    void cleanup();

    // Return a ROMaskedArray.  The new ROMaskedArray is masked by the input
    // LogicalArray "anded" with the mask of the original ROMaskedArray.
    // This mask must conform to the array, but it does not need to have the
    // same origin.
    ROMaskedArray<T> operator() (const LogicalArray &mask) const;

    // Return a ROMaskedArray.  The new ROMaskedArray is masked by the input
    // ROMaskedLogicalArray "anded" with the mask of the original ROMaskedArray.
    // This mask must conform to the array, but it does not need to have the
    // same origin.
    ROMaskedArray<T> operator() (const ROMaskedLogicalArray &mask) const;

    // Make a copy of the masked array.
    // This is a deep copy.  The Array and mask components of the returned
    // ROMaskedArray are deep copies of the Array and mask in the input
    // ROMaskedArray pointed to by this.  In other words, the Array and mask
    // in the output ROMaskedArray are completely independent of those in
    // the input ROMaskedArray.
    ROMaskedArray<T> copy() const;

    // Return the internal Array.
    const Array<T> & getArray() const;

    // Return the (const) internal Mask.
    const LogicalArray & getMask() const;

    // The dimensionality of this masked array.
    uInt ndim() const;

    // The number of elements of this masked array.
    // This is the number of elements of the mask which are TRUE.
    uInt nelements() const;

    // The number of elements in the underlying Array.
    uInt arrayNelements() const;

    // Check to see if the masked array is consistent. This is about the same
    // thing as checking for invariants. If AIPS_DEBUG is defined, this is
    // invoked after construction and on entry to most member functions.
    Bool ok() const;

    // Are the shapes identical? The origins do NOT need to be the same.
    // Binary operations will "line up" the masked arrays at their origins, so
    // as long as the shapes are the same the masked arrays conform.
    // <group>
    Bool conform(const Array<T> &other) const;
    Bool conform(const ROMaskedArray<T> &other) const;
    // </group>

    // The IPosition of the first element of the underlying Array.
    // NOTE: The mask origin is defined to be this also.
    IPosition origin() const;

    // The length of each axis.
    IPosition shape() const;


    // Manipulate the storage for the underlying Array.
    // See the description of the corresponding Array functions
    // for more information.
    // <group>
    const T * getArrayStorage (Bool &deleteIt) const;
    //
    void freeArrayStorage(const T *&storage, Bool deleteIt) const;
    // </group>


    // Manipulate the storage for the underlying Mask.
    // See the description of the corresponding Array functions
    // for more information.
    // <group>
    const LogicalArrayElem *getMaskStorage (Bool &deleteIt) const;
    //
    void freeMaskStorage(const LogicalArrayElem *&storage, Bool deleteIt) const;
    // </group>


protected:
    // The array.
    Array<T> *pArray;

    // The mask.
    LogicalArray *pMask;

    // Cache the number of elements.
    uInt nelem;

    // Is the number of elements cache valid?
    Bool nelemValid;


private:
    //# DO NOT define this.
    ROMaskedArray<T> &operator=(const ROMaskedArray<T> &other);
};


//# These typedefs seem to be necessary for instantiations of MaskedArrays
//# of these types.  Without them, instantiations of MaskedArrays of these
//# types cannot cause instantiations of their ROMaskedArray base class.
//# All of the types for which Arrays are instantiated are included.
//#
//# This is for the Sun cfront based C++ compiler.  It may be true
//# for other cfront based compilers as well.  It may not be necessary
//# for compilers using other technology.


typedef ROMaskedArray<char> ROMaskedArray_char;
typedef ROMaskedArray<unsigned char> ROMaskedArray_unsigned_char;
typedef ROMaskedArray<short> ROMaskedArray_short;
typedef ROMaskedArray<unsigned short> ROMaskedArray_unsigned_short;
typedef ROMaskedArray<int> ROMaskedArray_int;
typedef ROMaskedArray<unsigned int> ROMaskedArray_unsigned_int;
typedef ROMaskedArray<long> ROMaskedArray_long;
typedef ROMaskedArray<unsigned long> ROMaskedArray_unsigned_long;

typedef ROMaskedArray<float> ROMaskedArray_float;
typedef ROMaskedArray<double> ROMaskedArray_double;

typedef ROMaskedArray<Complex> ROMaskedArray_Complex;
typedef ROMaskedArray<DComplex> ROMaskedArray_DComplex;

typedef ROMaskedArray<String> ROMaskedArray_String;


// 
// <summary>
// Class for masking an Array for readwrite operations on that Array.
// </summary>
// <reviewed reviewer="" date="" tests="">
//
// <prerequisite>
//   <li> Array
//   <li> LogicalArray
//   <li> ROMaskedArray
// </prerequisite>
//
// <etymology>
// MaskedArray is a class for masking elements of an Array while performing
// operations on that Array.
// </etymology>
//
// <synopsis>
// A MaskedArray is an association between an Array and a mask.  The mask
// selects elements of the Array.  Only elements of the Array where the
// corresponding element of the mask is True are defined.  Thus, operations
// on a MaskedArray only operate on those elements of the Array where the 
// corresponding element of the mask is True.
//
// A MaskedArray should be thought of as a manipulator for an Array, analogous
// to an iterator.  It allows one to perform whole Array operations on selected
// elements of the Array.
//
// The mask used in the constructor for the MaskedArray must conform to
// the Array.  However, it does not need to have the same origin.
// The internal mask is (will be) copy constructed with reference semantics
// from the input mask.  Its origin is then shifted to have the same origin as
// the Array.  Therefore, it is (will be) possible to change the internal mask
// by changing values in the input mask *after* the MaskedArray has been
// constructed.  To ensure that the internal mask is independent of the input
// mask after construction, use mask.copy() as the input argument.
//
// One can explicitly construct a MaskedArray from an Array and a mask or
// a MaskedArray and a mask.  One can also use operator() on an Array or
// a MaskedArray to implicitly create a MaskedArray.
//
// One can create a MaskedArray from a MaskedArray and a mask.  The resulting
// MaskedArray has as its Array the Array from the original MaskedArray.
// The mask for the resulting MaskedArray is the AND of the mask from the
// original MaskedArray and the input mask.
//
// Any operation involving a MaskedArray or a set of MaskedArrays is only
// performed for those elements where the AND of the masks is True.
//
// Any operation involving a MaskedArray or a set of MaskedArrays results
// in a MaskedArray whose mask is the AND of the masks of the original
// MaskedArrays.  The only exception to this is assignment, where the
// mask determines which elements of the underlying Array are assigned.
//
// Masks, which are LogicalArrays, can be constructed by logical operations
// involving Arrays.  They can also, of course, be constructed by individually
// setting individual elements of an LogicalArray.
//
// Many mathematical and logical global operators and functions which operate
// on ROMaskedArrays and MaskedArrays are defined.  Typically, they are
// defined for all sensible combinations of "MaskedArrays", Arrays, and scalars.
// "MaskedArrays" means ROMaskedArrays if the underlying Array will not be
// modified, and means MaskedArrays if the underlying Array will be modified.
//
// Mathematical global operators and functions are defined in
// Arrays/MaskArrMath.h .
// The list is:
// <ul>
//   <li> operator+= ()
//   <li> operator-= ()
//   <li> operator*= ()
//   <li> operator/= ()
//   <li> operator+ ()
//   <li> operator- ()
//   <li> operator* ()
//   <li> operator/ ()
//   <li> sin ()
//   <li> cos ()
//   <li> tan ()
//   <li> asin ()
//   <li> acos ()
//   <li> atan ()
//   <li> sinh ()
//   <li> cosh ()
//   <li> tanh ()
//   <li> exp ()
//   <li> log ()
//   <li> log10 ()
//   <li> sqrt ()
//   <li> abs ()
//   <li> fabs ()
//   <li> ceil ()
//   <li> floor ()
//   <li> atan2 ()
//   <li> fmod ()
//   <li> pow ()
//   <li> minMax ()
//   <li> min ()
//   <li> max ()
//   <li> indgen ()
//   <li> sum ()
//   <li> sumsquares ()
//   <li> product ()
//   <li> mean ()
//   <li> variance ()
//   <li> stddev ()
//   <li> avdev ()
//   <li> median ()
//   <li> square ()
//   <li> cube ()
// </ul>
//
// Logical global operators and functions are defined in
// Arrays/MaskArrLogi.h .
// The list is:
// <ul>
//   <li> allLE ()
//   <li> allLT ()
//   <li> allGE ()
//   <li> allGT ()
//   <li> allEQ ()
//   <li> allNE ()
//   <li> allAND ()
//   <li> allOR ()
//   <li> anyLE ()
//   <li> anyLT ()
//   <li> anyGE ()
//   <li> anyGT ()
//   <li> anyEQ ()
//   <li> anyNE ()
//   <li> anyAND ()
//   <li> anyOR ()
//   <li> operator<= ()
//   <li> operator<  ()
//   <li> operator>= ()
//   <li> operator<  ()
//   <li> operator== ()
//   <li> operator!= ()
//   <li> operator&& ()
//   <li> operator|| ()
// </ul>
// </synopsis>
//
// <example>
// Use an explicit MaskedArray to limit the maximum value of an Array.
//
// <srcblock>
//   Array<Int> arr (20);
//      . . .
//   MaskedArray<Int> marr (arr, (arr > 5));
//   marr = 5;
// </srcblock>
//
// This sets all elements of arr which are greater than 5 to 5.
// </example>
//
// <example>
// Use an implicit MaskedArray to limit the minimum value of an Array.
//
// <srcblock>
//   Array<Int> arr (20);
//      . . .
//   arr (arr < 0) = 0;
// </srcblock>
//
// This sets all elements of arr which are less than 0 to 0.
// </example>
//
// <example>
// It does not matter where in an expression the MaskedArrays are located.
// The operation is only performed on those elements where the AND of the
// masks is True.
//
// The following expressions are all equivalent.
// The first (and second) expressions are the most efficient, since the sum
// is only performed for those elements where ((a > 0) && (b > 0)).
// The third example is less efficient, since the sum is performed for
// all elements of a and b, and then the assignment is only performed for
// those elements where ((a > 0) && (b > 0)).
//
// <srcblock>
//   Array<Int> arr (20);
//   Array<Int> a (20);
//   Array<Int> b (20);
//      . . .
//   arr = a(a > 0) + b(b > 0);
//
//   arr = a(b > 0) + b(a > 0);
//
//   arr ((a > 0) && (b > 0)) = a + b;
//
//   arr = (a + b) ((a > 0) && (b > 0));
//
//   arr (a > 0) = a + b(b > 0);
//
// </srcblock>
//
// All of these expressions set those elements of arr where
// ((a > 0) && (b > 0)) to a + b.  Those elements of arr where the condition
// is False are unchanged.
// </example>
//
// <motivation>
// A MaskedArray is an association between an Array and a LogicalArray which
// masks the Array.  It allows one to perform whole Array manipulations
// with a single expression, selecting those elements of the Array to modify
// based either on a logical expression, typically involving some of the
// Arrays involved in the expression, or based on a specifically set mask.
// </motivation>
//
// <todo asof="$DATE:$>
//   <li> Consider whether there should be constructors for masks
//          specified as Array<Bool>.
//   <li> Change constructors to always do copy construction with
//          reference semantics when creating the internal mask.
//          Then shift the internal mask's origin to coincide with
//          that of the Array.
//
//          The Array class must have a "shiftOrigin()" member function
//          added so that this can work.
// </todo>


template<class T> class MaskedArray : public ROMaskedArray<T>
{

public:

    // Create a MaskedArray from an Array and a LogicalArray.
    MaskedArray(const Array<T> &inarray, const LogicalArray &inmask);

    // Create a MaskedArray from a MaskedArray and a LogicalArray.
    MaskedArray(const MaskedArray<T> &inarray, const LogicalArray &inmask);

    // Create a MaskedArray from an Array and a ROMaskedLogicalArray.
    MaskedArray(const Array<T> &inarray, const ROMaskedLogicalArray &inmask);

    // Create a MaskedArray from a MaskedArray and a ROMaskedLogicalArray.
    MaskedArray(const MaskedArray<T> &inarray,
                const ROMaskedLogicalArray &inmask);

    MaskedArray(const MaskedArray<T> &other);

    // Frees up storage.
    ~MaskedArray();

    // This function is used by the exception handling mechanism we have
    // defined. It merely calls the destructor. When real exceptions are
    // available it will be unnecessary.
    //*display 8
    void cleanup();

    // Copy the values in inarray to this, only copying those elements
    // for which the corresponding mask element is True.
    //
    // <thrown>
    //    <li> ArrayConformanceError
    // </thrown>
    //
    MaskedArray<T> &operator=(const Array<T> &inarray);

    // Copy the values in other to this, only copying those elements
    // for which the logical AND of the corresponding mask elements
    // of both MaskedArrays is True.
    //
    // <thrown>
    //    <li> ArrayConformanceError
    // </thrown>
    //
    MaskedArray<T> &operator=(const ROMaskedArray<T> &other);

    // Set every element of this array to "value", only setting those elements
    // for which the corresponding mask element is True.
    // In other words, a scalar behaves as if it were a constant conformant
    // array.
    MaskedArray<T> &operator=(const T &value);

    // Return a ROMaskedArray.  The new ROMaskedArray is masked by the input
    // LogicalArray "anded" with the mask of the original ROMaskedArray.
    // This mask must conform to the array, but it does not need to have the
    // same origin.
    ROMaskedArray<T> operator() (const LogicalArray &mask) const;

    // Return a MaskedArray.  The new MaskedArray is masked by the input
    // LogicalArray "anded" with the mask of the original MaskedArray.
    // This mask must conform to the array, but it does not need to have the
    // same origin.
    MaskedArray<T> operator() (const LogicalArray &mask);

    // Return a ROMaskedArray.  The new ROMaskedArray is masked by the input
    // ROMaskedLogicalArray "anded" with the mask of the original ROMaskedArray.
    // This mask must conform to the array, but it does not need to have the
    // same origin.
    ROMaskedArray<T> operator() (const ROMaskedLogicalArray &mask) const;

    // Return a MaskedArray.  The new MaskedArray is masked by the input
    // ROMaskedLogicalArray "anded" with the mask of the original MaskedArray.
    // This mask must conform to the array, but it does not need to have the
    // same origin.
    MaskedArray<T> operator() (const ROMaskedLogicalArray &mask);

    // Return the internal Array, writeable.
    Array<T> & getRWArray() const;

    // Manipulate the storage for the underlying Array.
    // See the description of the corresponding Array functions
    // for more information.
    // <group>
    T * getRWArrayStorage (Bool &deleteIt) const;
    //
    void putArrayStorage(T *&storage, Bool deleteAndCopy) const;
    // </group>

};


// 
// Dummy function to simulate section introduction documentation.
// <summary>
// General global functions for MaskedArrays.
// </summary>
// <reviewed reviewer="" date="" tests="">
//
// <prerequisite>
//   <li> Array
//   <li> LogicalArray
//   <li> MaskedArray
// </prerequisite>
//
// <synopsis>
// These are generally useful global functions which operate on all
// MaskedArrays.
// </synopsis>
void AAA_MaskedArray_Global_Functions_Introduction ();


// Test conformance for masked arrays and arrays of different types.
// Are the shapes identical? The origins do NOT need to be the same.
// Binary operations will "line up" the masked arrays at their origins, so
// as long as the shapes are the same the masked arrays conform.
//
// <group>
//
template<class T, class U>
  Bool conform2 (const ROMaskedArray<T> &left, const Array<U> &right);
template<class T, class U>
  Bool conform2 (const Array<T> &left, const ROMaskedArray<U> &right);
template<class T, class U>
  Bool conform2 (const ROMaskedArray<T> &left, const ROMaskedArray<U> &right);
//
// </group>

// Make a writeable copy of the masked array.
// This is a deep copy.  The Array and mask components of the returned
// MaskedArray are copies of the Array and mask in the input
// ROMaskedArray.  In other words, the Array and mask
// in the output MaskedArray are completely independent of those in
// the input ROMaskedArray.
template<class T> MaskedArray<T> RWcopy (const ROMaskedArray<T> &marray);

#endif
