//# Tables.h: The Tables module - AIPS++ data storage
//# Copyright (C) 1994
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: aips2-request@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id: Tables.h,v 1.1.1.1 1996/04/04 16:37:03 dschieb Exp $

#if !defined(AIPS_TABLES_H)
#define AIPS_TABLES_H

//# Includes
#include <aips/Tables/TableDesc.h>
#include <aips/Tables/ColumnDesc.h>
#include <aips/Tables/ScaColDesc.h>
#include <aips/Tables/ArrColDesc.h>
#include <aips/Tables/SubTabDesc.h>

#include <aips/Tables/StManAipsIO.h>

#include <aips/Tables/Table.h>
#include <aips/Tables/SetupNewTab.h>
#include <aips/Tables/ScalarColumn.h>
#include <aips/Tables/ArrayColumn.h>
#include <aips/Array.h>
#include <aips/Lattice/Slicer.h>
#include <aips/Slice.h>

#include <aips/Tables/ExprNode.h>

#include <aips/Tables/TableVector.h>
#include <aips/Tables/TabVecMath.h>
#include <aips/Tables/TabVecLogic.h>


// <module>
// 
// <summary> Tables are the data storage mechanism for AIPS++ </summary>

// <reviewed reviewer="jhorstko" date="1994/08/30" tests="" demos="">

// <prerequisite>
//      <li> Keywords
// </prerequisite>
// 
// <etymology>
// "Table" is a formal term from relational database theory: 
//   'The organizing principle in a relational database is the TABLE,
//    a rectangular, row/column arrangement of data values.'
// AIPS++ tables are extensions to traditional tables, but are similar
// enough that we use the same name.  There is also a strong resemblance
// between the uses of AIPS++ tables, and FITS binary tables, which
// provides another reason to use "Tables" to describe the AIPS++ data
// storage mechanism.
// </etymology>
//
// <synopsis> 
// Tables are the fundamental storage mechanism for AIPS++.  They are
// based upon the ideas of Allen Farris, as laid out in <url needed here>.
// 
// More detailed information can be obtained in the AIPS++ "Database" 
// document, found
// <a href="http://www.cv.nrao.edu:85/info/Database"><em>here</em></a>.
// 
// Traditional relational database tables have two features that
// decisively limit their applicability to scientific data.  First, an item of
// data in a column of a table must be atomic -- it must have no internal
// structure.  A consequence of this restriction is that relational
// databases are unable to deal with arrays of data items.  Second, an
// item of data in a column of a table must not have any direct or
// implied linkages to other items of data or data aggregates.  This
// restriction makes it difficult to model complex relationships between
// collections of data.  While these restrictions may make it easy to
// define a mathematically complete set of data manipulation operations,
// they are simply intolerable in a scientific data-handling context.
// Multi-dimensional arrays are frequently the most natural modes in
// which to discuss and think about scientific data.  In addition,
// scientific data often requires complex calibration operations that
// must draw on large bodies of data about equipment and its performance
// in various states.  The restrictions imposed by the relational model
// make it very difficult to deal with complex problems of this nature.
// 
// In response to these limitations, and other needs, AIPS++ tables have
// the following features:
// <ul>
//  <li> Any number of keyword/value pairs. Keywords may be
//         attached to the table as a whole, or to any column.
//         They may be freely added, retrieved, re-assigned, or deleted.
//         They are, in essence, a self-resizing list of values (any
//         of the primitive types) indexed by Strings (the keyword name).
//         A table keyword/value pair might be
//              Observer = Grote Reber
//              Date = 10 october 1942
//         Column keyword/value pairs might be
//              Units =  mJy
//              Reference Pixel = 320
//  <li> Columns may contain actual data, or may simply indicate how
//         the data will be generated dynamically.  The first of these is
//         called a "filled column"  and the second a "virtual column".
//  <li> A cell in a column may contain a scalar, a "direct" array (which
//         must have the same shape in all cells of a column, are usually small,
//         and are stored), an "indirect" array (which may have varying shapes,
//         in different cells of the same column, may be arbitrarily large,
//         and which are stored in separate files).
//         A cell in a column may also contain another table, allowing
//         a table to express arbitrarily complex data relationships.
//         For example....
//  <li> The table may use more than one "storage manager".  These
//         are software entities that store data, typically to disk.
//         Tables are designed so that each column can use a different
//         storage manager, making it possible to choose a storage technique
//         optimized for the kind of data stored in a particular column.  
//  <li> "Virtual column engines" are used to implement the "virtual"
//         or "calculated-on-the-fly" columns mentioned above.
//         The Table system provides an abstract base class (or "interface
//         class") specifying the protocol for these engines.  The
//         programmer must derive a concrete class to implement
//         the application-specific virtual column.  For example:
//         the programmer needs a column in a table which is the
//         difference between two other columns.  (Perhaps these
//         two other columns are updated periodically during the
//         execution of a program...)  A good way to handle this would
//         be to attach a virtual column to the table, and write
//         a virtual column engine which knows to calculate the difference
//         between corresponding cells of the two other columns.
//         So the result is that accessing a particular cell of the
//         virtual column invokes the virtual column engine, which
//         then gets the values from the other two columns, and
//         returns their difference.
//  <li> Only the standard AIPS++ data types can be used in filled
//         columns, be they scalars or arrays.  These are:  Bool, uChar,
//         Int, uInt, Float, Double, Complex, DComplex and String.
//  <li> The table contains a description of itself, which may also
//         exist in a separate file.  This description must be created
//         as the first step in creating a new table; it specifies the
//         number and kind of columns, and the number and kinds of
//         keywords.
//  <li> If you need to create a number of tables which all have the same
//         structure, then this separate table description file might be useful.
//         In other circumstances it probably should be avoided, and the
//         only description should be the one built-in to the table
//         itself.
//  <li> Initial sets of keywords for the table and columns may be
//         defined in the table description.
//  <li> A column can have a default value, which can automatically be stored
//         in a cell in the table column, when a row is added to the table.
// </ul> 
//
//  <h3> Table Keywords </h3>
// The class <linkto class=TableKeywords><src>TableKeywords</src></linkto> 
// (<linkto class="TableKeywords:description"><em>description</em></linkto>)represents
// the keywords in a table.  It is (indirectly) derived from the standard 
// keyword class <src>KeywordSet</src>. It is possible to attach keywords to 
// the table and to each individual column. These keywords can be defined or
// removed freely.
// 

//  <h3> Table Description </h3>
// The table description defines the layout of the columns in the table.
// Furthermore it is possible to define the initial sets of keywords
// for the table and columns.
// A column can have a default value, which can automatically be stored
// in a cell in the table column, when a row is added to the table.
//
// The public classes to set up a table description are:
// <ol>
//  <li> <linkto class=TableDesc>TableDesc</linkto>
//   holds the table description (<linkto class="TableDesc#important"><em>info</em></linkto>)
//  <li> <linkto class=ColumnDesc>ColumnDesc</linkto>
//   holds a generic column description. (<linkto file="ColumnDesc.h#info"><em>info</em></linkto>)
//  <li> <linkto class=ScalarColumnDesc>ScalarColumnDesc</linkto>
//   defines a column containing a scalar value.
//  <li> <linkto class=ArrayColumnDesc>ArrayColumnDesc</linkto>
//   defines a column containing an (in)direct array.
//  <li> <linkto class=SubTableDesc>SubTableDesc</linkto>
//   defines a column containing a table.
// </ol>
//
// The header files of the above mentioned classes contain more
// information on how to use them.

// The following example how a table description can be constructed.
// <srcblock>
// #include <aips/Tables/TableDesc.h>
// #include <aips/Tables/ScaColDesc.h>
// #include <aips/Tables/ArrColDesc.h>
// #include <aips/Tables/SubTabDesc.h>
// #include <aips/ScalarKeySet.h>
// #include <aips/IPosition.h>
// #include <aips/Vector.h>
//
// main()
// {
//     // First build the new description of a subtable.
//     // Define keyword subkey (integer) having value 10.
//     // Define columns ra and dec (double).
//     TableDesc subTableDesc("tTableDesc_sub", "1", TableDesc::New);
//     subTableDesc.keywordSet().keysInt()("subkey") = 10;
//     subTableDesc.addColumn (ScalarColumnDesc<double> ("ra"));
//     subTableDesc.addColumn (ScalarColumnDesc<double> ("dec"));
//
//     // Now create a new table description
//     // Define a comment for the table description.
//     // Define some keywords.
//     ColumnDesc colDesc1, colDesc2;
//     TableDesc td("tTableDesc", "1", TableDesc::New);
//     td.comment() = "A test of class TableDesc";
//     td.keywordSet().keysfloat()("ra") = 3.14;
//     td.keywordSet().keysdouble()("equinox") = 1950;
//     td.keywordSet().keysInt()("aa") = 1;
//
//     // Define an integer column ab.
//     td.addColumn (ScalarColumnDesc<Int> ("ab", "Comment for column ab"));
//
//     // Add a scalar integer column ac, define keywords for it
//     // and define a default value 0.
//     // Overwrite the value of keyword unit.
//     ScalarColumnDesc<Int> acColumn("ac");
//     acColumn.keywordSet().keysComplex()("scale") = 0;
//     acColumn.keywordSet().keysString()("unit") = "";
//     acColumn.setDefault (0);
//     td.addColumn (acColumn);
//     td["ac"].keywordSet().keysString()("unit") = "DEG";
//
//     // Add a scalar string column ad and define its comment string.
//     td.addColumn (ScalarColumnDesc<String> ("ad","comment for ad"));
//
//     // Now define array columns.
//     // This one is indirect and has no dimensionality mentioned yet.
//     td.addColumn (ArrayColumnDesc<Complex> ("Arr1","comment for Arr1"));
//     // This one is indirect and has 3-dim arrays.
//     td.addColumn (ArrayColumnDesc<Int> ("A2r1","comment for Arr1",3));
//     // This one is direct and has 2-dim arrays with axes length 4 and 7.
//     td.addColumn (ArrayColumnDesc<uInt> ("Arr3","comment for Arr1",
//        				    IPosition(2,4,7),
// 					    ColumnDesc::Direct));
//
//     // Add columns containing tables.
//     // This is done in 3 slighty different ways, which all have
//     // their own (dis)advantages.
//     // See SubTabDesc.h for a description of the SubTableDesc constructors.
//     td.addColumn (SubTableDesc("sub1", "subtable by name","tTableDesc_sub"));
//     td.addColumn (SubTableDesc("sub2", "subtable copy",    subTableDesc));
//     td.addColumn (SubTableDesc("sub3", "subtable pointer", &subTableDesc));
// }
// </srcblock>
//
// This example only shows the basic table description functionality.
// More specialized things, like defining a default data manager
// are not shown.
// 
//
//  <h3> Creating Tables </h3>
// Once a table description has been created, it can be used to generate
// a table from it. This is a multi-step process:
// <ol>
//  <li>
//   Create an object SetupNewTable with the name of the new table.
//  <li>
//   Define data managers as needed. If this is not done, data
//   managers will be defined as needed using the default data
//   manager name and group in the column descriptions.
//  <li>
//   Bind columns to the appropriate data manager (i.e. a storage
//   manager or a virtual column engine).
//  <li>
//   Define the shape of direct columns (if not already defined in the
//   column description).
//  <li>
//   Create the Table object from the SetupNewTable object. This final
//   step will perform a check and will create the files as needed.
// </ol>
//
// Not all data managers support all the table functionality.
// For example, deleting a column or row is not supported in the
// Karma storage manager.
// Currently no virtual column engines are implemented yet.
// Two storage managers are supported:
// <ol>
//  <li>
//   <linkto class=StManAipsIO>StManAipsIO</linkto> uses AipsIO
//   to store the data in the columns.
//   It supports are table functionality, but its IO is probably not
//   as efficient as other storage managers. It also requires that
//   a large part of the table fits in memory.
//  <li>
//   <linkto class=StManKarma>StManKarma</linkto> uses Karma to store the data in the columns.
//   (Karma is a software package developed by Richard Gooch at CSIRO).
//   Currently Karma can only handle scalars and direct arrays.
//   Addition and deletion of rows and columns is not supported.
//   The table size (i.e. the number of rows) has to be defined
//   when the Table object is created from the SetupNewTable object.
// </ol>
// It should be clear that the choice of data manager greatly
// influences the type of operations you can do on a table.
// For example, if only one column uses the Karma storage manager,
// it is not possible to add or delete rows. However, adding a
// column is still possible by using an AipsIO storage manager
// for the column to be added.
//
// The following example shows how a table can be created.
// The next section discusses how to access the table.
// <srcblock>
// #include <aips/Tables/TableDesc.h>
// #include <aips/Tables/SetupNewTab.h>
// #include <aips/Tables/Table.h>
// #include <aips/Tables/ScaColDesc.h>
// #include <aips/Tables/ArrColDesc.h>
// #include <aips/Tables/StManAipsIO.h>
// #include <aips/Tables/StManKarma.h>
// 
// main()
// {
//     // First build the table description.
//     TableDesc td("tTableDesc", "1", TableDesc::Scratch);
//     td.comment() = "A test of class SetupNewTable";
//     td.addColumn (ScalarColumnDesc<Int> ("ab" ,"Comment for column ab"));
//     td.addColumn (ScalarColumnDesc<Int> ("ac"));
//     td.addColumn (ScalarColumnDesc<uInt> ("ad","comment for ad"));
//     td.addColumn (ScalarColumnDesc<float> ("ae"));
//     td.addColumn (ArrayColumnDesc<float> ("arr1",3,ColumnDesc::Direct));
//     td.addColumn (ArrayColumnDesc<float> ("arr2",0));
//     td.addColumn (ArrayColumnDesc<float> ("arr3",0,ColumnDesc::Direct));
// 
//     // Setup a new table from the description.
//     SetupNewTable newtab("newtab.data", td, Table::New);
//     // Create storage managers for it.
//     StManAipsIO stmanAipsIO_1 ();
//     StManAipsIO stmanAipsIO_2 ();
//     StManKarma  stmanKarma    ();
// 
//     // Start with binding all columns to the first storage manager.
//     newtab.bindAll (stmanAipsIO_1);
//     // Bind a few columns to another storage manager.
//     newtab.bindColumn ("ab", stmanAipsIO_2);
//     newtab.bindColumn ("ae", stmanKarma);
//     newtab.bindColumn ("arr3", stmanKarma);
// 
//     // Define the shape of the direct columns.
//     // (this could have been defined in the column description).
//     newtab.setShapeColumn( "arr1", IPosition(3,2,3,4));
//     newtab.setShapeColumn( "arr3", IPosition(3,3,4,5));
// 
//     // Finally create the table consisting of 10 rows.
//     // Defining the number of rows is necessary, because the
//     // Karma storage manager is used.
//     Table tab(newtab, 10);
// 
//     // Now we can fill the table, which is shown in a next section.
//     // The Table destructor will flush the table to the files.
// }
// </srcblock>
// 
//
//  <h3> Opening an Existing Table </h3>
// An existing table can be opened by creating a Table for it.
// The constructor option determines if the table is opened as
// readonly or as read/write. A readonly table file must be opened 
// as readonly, otherwise an exception is thrown.
//
// The functions isWritable can be used to determine if a table
// is writable.
//
// When the table is opened, the data managers will be reinstantiated
// as defined when the table was created.
//
// Opening an existing table is straightforward and looks like:
// <srcblock>
//     Table table ("tableName");                     // readonly
// </srcblock>
// or
// <srcblock>
//     Table table ("tableName", Table::Update);      // read/write
// </srcblock>
// 
//
//  <h3> Writing into a Table </h3>
// Once a table has been created or has been opened for read/write,
// data needs to be written into the table. Before that can be done,
// it may be needed to add one or more rows to the table.
// <note role=tip> If a table is created with a given number of rows, it
// is not needed to add rows.
//
// When rows are added to the table (either via the Table constructor
// or via the addRow function), it is possible to define if the
// rows need to be initialized. Initializing means that the default
// value as defined in the description gets written into all cells
// of the added rows.
//
// The actual writing of data must be done using the classes
// <linkto class=ScalarColumn>ScalarColumn</linkto> and 
// <linkto class=ArrayColumn>ArrayColumn</linkto>. For each column one or more of
// these objects can be constructed. The put functions in these classes
// allow to write a value at a time or the entire column in one go.
// For arrays it is possible to put subsections of the arrays.
//
// A typical program could look like:
// <srcblock>
// #include <aips/Tables/TableDesc.h>
// #include <aips/Tables/SetupNewTab.h>
// #include <aips/Tables/Table.h>
// #include <aips/Tables/ScaColDesc.h>
// #include <aips/Tables/ArrColDesc.h>
// #include <aips/Tables/ScalarColumn.h>
// #include <aips/Tables/ArrayColumn.h>
// #include <aips/Vector.h>
// #include <aips/Lattice/Slicer.h>
// #include <aips/ArrayMath.h>
// #include <iostream.h>
// 
// main()
// {
//     // First build the table description.
//     TableDesc td("tTableDesc", "1", TableDesc::Scratch);
//     td.comment() = "A test of class SetupNewTable";
//     td.addColumn (ScalarColumnDesc<Int> ("ac"));
//     td.addColumn (ArrayColumnDesc<float> ("arr2",0));
// 
//     // Setup a new table from the description.
//     // Create the (still empty) table.
//     // Note that since no explicit binding of columns to data managers
//     // has been done, implicit binding to the default AipsIO storage
//     // manager will be done.
//     SetupNewTable newtab("newtab.data", td, Table::New);
//     Table tab(newtab);
//
//     // Now construct the various column objects.
//     // Their data type has to match the data type as in the description.
//     ScalarColumn<Int> ac (tab, "ac");
//     ArrayColumn<float> arr2 (tab, "arr2");
//     Vector<float> vec2(100);
//
//     // Now write the data into the columns.
//     // In each cell arr2 will be a vector of length 100.
//     // Since its shape is not explicitly set, it is set implicitly.
//     for (uInt i=0; i<10; i++) {
//         tab.addRow();               // First add a row.
//         ac.put (i, i+10);           // value is i+10 in row i
//         indgen (vec2, float(i+20)); // vec2 gets i+20, i+21, ..., i+119
//         arr2.put (i, vec2); 
//     }
//
//     // Now show the entire column ac.
//     // Show the 10th element of arr2.
//     cout << ac.getColumn();
//     cout << arr2.getColumn (Slicer(Slice(10)));
//
//     // The Table destructor writes the table.
// }
// </srcblock>
//
// In this example rows are added in the for loop.
// It would have also been possible to immediately create 10 rows
// by constructing the Table object as:
// <srcblock>
//     Table tab(newtab, 10);
// </srcblock>
// In that case the function call
// <srcblock>
//     tab.addRow()
// </srcblock>
// should not be done.
//
//
// In the classes <linkto class=TableColumn>TableColumn</linkto>, 
// <linkto class=ScalarColumn>ScalarColumn</linkto> and 
// <linkto class=ArrayColumn>ArrayColumn</linkto> are
// several functions which can be used to put a value into a cell of
// the column or to put all values in the column. At first this may
// look overwhelming, but it is quite simple.
// The functions can be divided in a two cases:
// <ol>
//  <li>
//   Put the given value into the column cell(s).
//   <ul>
//    <li>
//     The simplest put function in 
//     <linkto class=ScalarColumn>Scalar</linkto>/<linkto class=ArrayColumn>ArrayColumn</linkto> 
//     puts a value into
//     the given column cell. For convenience, there is a putSlice
//     function in ArrayColumn which puts only a part of the array.
//    <li>
//     The fillColumn function in Scalar/ArrayColumn fills an
//     entire column by putting the given value into all column cells.
//    <li>
//     The simplest putColumn function in Scalar/ArrayColumn puts an
//     array of values into the column.
//     For convenience, there is a putColumn function in ArrayColumn
//     which puts only a part of the arrays.
//   </ul>
//  <li>
//   The other put and putColumn functions copy values from another
//   column to this column. These functions have the advantage that the
//   data type of the input and/or output column can be unknown.
//   The generic (RO)TableColumn objects can be used for this purpose.
//   The put(Column) function takes care of checking and, if possible,
//   converting the data types. If the conversion is not possible, an
//   exception will be thrown.
//   <ul>
//    <li>
//     The put functions copy the value in a cell of the input column
//     to a cell in the output column. The row numbers of the cells
//     in the columns can be different.
//    <li>
//     The putColumn functions copy the entire contents of the input column
//     to the output column. The lengths of the columns must be equal.
//   </ul>
//   Each class has its own set of these functions.
//   <ul>
//    <li>
//     TableColumn has the most generic put/putColumn functions.
//     They can be used if the data types of both input and output
//     column are unknown.
//     Note that these functions are virtual.
//    <li>
//     The put/putColumn functions in ScalarColumn and ArrayColumn are
//     less generic and therefore potentially more efficient.
//     The most efficient variants are the ones taking a <src>ROScalar/
//     ArrayColumn<T></src>, because they require no data type
//     conversion.
//   </ul>
// </ol>
// 
//
//  <h3> Reading from a Table </h3>
// Reading from a column in a table can be done using the get functions
// in the classes <src>ROScalarColumn<T></src> and ROArrayColumn<T>.
// Apart from these functions, ROTableColumn::getScalar and
// ROTableColumn:asXXX can
// be used to get scalars of a standard data type (i.e. Bool, uChar, Int,
// uInt, float, double, Complex, DComplex and String). These functions
// also offer the possibility of data type promotion,
// thus getting a value in a float column as a double, for example.
//
// The way these functions are used is the same as the simple
// put functions described in the previous section.
//
// ScalarColumn<T> is derived from ROScalarColumn<T>, thus these get
// functions are also available in ScalarColumn<T>. However, if
// ScalarColumn<T> object is constructed for a non-writable column,
// an exception is thrown. Only ROScalarColumn<T> objects can be
// constructed for nonwritable columns.
// The same thing is true for ArrayColumn<T> and TableColumn.
// 
//
//  <h3> Table Selecting and Sorting </h3>
// The result of a select and sort of a table is another table,
// which references the original table. This means that an update
// of a sorted or selected table results in the update of the original
// table.
//
// It is possible to select rows or columns from a table.
// Columns can be selected by the Table::project function.
// Rows can be selected by the various Table function operators.
// The main way to do a row select is by giving a select expression
// using TableExprNode objects. These objects represent the various
// node in an expression (e.g. a constant, column, subexpression).
// The Table function col creates a TableExprNode object for a
// column. The function key does the same for a keyword by reading
// the keyword value and storing it as a constant in an expression node.
// All column nodes in an expression must belong to the same table,
// otherwise an exception is thrown.
// The following example selects all rows with RA>10
// <srcblock>
//    #include <aips/Tables/ExprNode.h>
//        ...
//    Table result = table.select (table.col("RA") > 10);
// </srcblock>
// while the following example selects between the given RA and DEC values.
// <srcblock>
//    Table result = table.select (table.col("RA") > 10
//                              && table.col("RA") < 14
//                              && table.col("DEC") >= -10
//                              && table.col("DEC") <= 10);
// </srcblock>
// The following operators can be used to form an arbitrarily
// complex expression:
// <ul>
//  <li> Relational operators ==, !=, > >= < and <=.
//  <li> Logical operators &&, || and !.
//  <li> Arithmetic operators +, -, *, / and unary + and -.
// </ul>
// In the near future functions like sin will also be possible.
// 
//
//  <h3> Table Vectors </h3>
// 
//
//  <h3> Table Iterators </h3>
// 
// </module>

#endif
