//# StManKarma.h: Storage manager for tables using Karma
//# Copyright (C) 1994
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: aips2-request@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id: StManKarma.h,v 1.1.1.1 1996/04/04 16:37:09 dschieb Exp $

#if defined(_AIX)
#pragma implementation ("../../implement/Tables/StManKarma.cc")
#endif 

#if !defined(AIPS_STMANKARMA_H)
#define AIPS_STMANKARMA_H

//# Includes
#include <aips/aips.h>
#include <aips/Tables/StMan.h>
#include <aips/Block.h>
#include <aips/Complex.h>
#include <aips/IPosition.h>
#include <aips/String.h>
#include <karma_storage.h>

//# Forward Declarations
class StManKarma;

//*display 4
// 
// <summary> Karma table column storage manager class </summary>
// <reviewed reviewer="" date="" tests="">
//
// <prerequisite>
//   <li> StManColumn
//   <li> StManKarma
// </prerequisite>
//
// <etymology>
// StManColumnKarma handles a column for a Karma storage manager.
// </etymology>
//
// <synopsis> 
// StManColumnKarma is used by StManKarma to handle the access to
// the scalar data in a table column.
// It is an storage manager based on Karma. It uses the Karma
// storage package to handle the data.
// It supports all standard data types except String.
// Addition and removal of rows are not supported. The exact
// number of rows have to be allocated at table creation time.
//
// The scalars are represented in Karma as a 1-D array of packets.
// A packet contains the scalar values for a particular row.
// </synopsis> 
//
// <motivation>
// StManColumnKarma handles the standard data types. The class
// is not templated, but a switch statement is used instead.
// Templates would cause too many instantiations.
// </motivation>
//
// <todo asof="$DATE:$">
//   <li> Implement get/putBlock.
// </todo>

class StManColumnKarma : public StManColumn
{
public:

    // Create a column of the given name and type.
    // The name is used as the Karma name of the column. If the column
    // gets renamed in the table, this name will NOT change.
    StManColumnKarma (StManKarma*, const String& name, int dataType);

    // Frees up the storage.
    ~StManColumnKarma();

    // Create the Karma description for this scalar column.
    // Only the *sca* arguments are used and filled in.
    virtual void doCreate (uInt& nrsca, unsigned int& scaType, char*& scaDesc,
			   uInt& nrarr, unsigned int& arrType, char*& arrDesc);

    // Convert data type from Aips++ to Karma.
    unsigned int convertDataType (int dataType) const;

    // Get a scalar value in the given row.
    // <group>
    void get (uInt rownr, Bool* dataPtr);
    void get (uInt rownr, uChar* dataPtr);
    void get (uInt rownr, Int* dataPtr);
    void get (uInt rownr, uInt* dataPtr);
    void get (uInt rownr, float* dataPtr);
    void get (uInt rownr, double* dataPtr);
    void get (uInt rownr, Complex* dataPtr);
    void get (uInt rownr, DComplex* dataPtr);
    void get (uInt rownr, String* dataPtr);
    // </group>

    // Put a scalar value into the given row.
    // <group>
    void put (uInt rownr, const Bool* dataPtr);
    void put (uInt rownr, const uChar* dataPtr);
    void put (uInt rownr, const Int* dataPtr);
    void put (uInt rownr, const uInt* dataPtr);
    void put (uInt rownr, const float* dataPtr);
    void put (uInt rownr, const double* dataPtr);
    void put (uInt rownr, const Complex* dataPtr);
    void put (uInt rownr, const DComplex* dataPtr);
    void put (uInt rownr, const String* dataPtr);
    // </group>

    // Get values from the given row on with a maximum of nrmax values.
    // This is used to get an entire column of scalars.
    // The buffer pointed to by dataPtr has to have the correct length
    // (which is guaranteed by the ScalarColumn getColumn function).
    // This is not implemented yet.
    // <group>
    uInt getBlock (uInt rownr, uInt nrmax, Bool* dataPtr);
    uInt getBlock (uInt rownr, uInt nrmax, uChar* dataPtr);
    uInt getBlock (uInt rownr, uInt nrmax, Int* dataPtr);
    uInt getBlock (uInt rownr, uInt nrmax, uInt* dataPtr);
    uInt getBlock (uInt rownr, uInt nrmax, float* dataPtr);
    uInt getBlock (uInt rownr, uInt nrmax, double* dataPtr);
    uInt getBlock (uInt rownr, uInt nrmax, Complex* dataPtr);
    uInt getBlock (uInt rownr, uInt nrmax, DComplex* dataPtr);
    uInt getBlock (uInt rownr, uInt nrmax, String* dataPtr);
    // </group>

    // Put nrmax values from the given row on.
    // This is used to put an entire column of scalars.
    // The buffer pointed to by dataPtr has to have the correct length
    // (which is guaranteed by the ScalarColumn putColumn function).
    // This is not implemented yet.
    // <group>
    uInt putBlock (uInt rownr, uInt nrmax, const Bool* dataPtr);
    uInt putBlock (uInt rownr, uInt nrmax, const uChar* dataPtr);
    uInt putBlock (uInt rownr, uInt nrmax, const Int* dataPtr);
    uInt putBlock (uInt rownr, uInt nrmax, const uInt* dataPtr);
    uInt putBlock (uInt rownr, uInt nrmax, const float* dataPtr);
    uInt putBlock (uInt rownr, uInt nrmax, const double* dataPtr);
    uInt putBlock (uInt rownr, uInt nrmax, const Complex* dataPtr);
    uInt putBlock (uInt rownr, uInt nrmax, const DComplex* dataPtr);
    uInt putBlock (uInt rownr, uInt nrmax, const String* dataPtr);
    // </group>

    // Check if the class invariants still hold.
    virtual Bool ok() const;

protected:
    // The storage manager.
    StManKarma* stmanPtr_p;
    // The name of the column.
    String name_p;
    // The data type of the data (as defined in DataType.h).
    int   dtype_p;
    // The data type of the data (as defined in Karma).
    unsigned int   karmaDtype_p;
    // A pointer to the rownr coordinate name.
    // It is a pointer to the coordinate name in the parent StManKarma object.
    const char* rowCoordPtr_p;

    // Fill in Karma element type and descriptor.
    void fillElem (unsigned int& elemType, char*& elemDesc);

private:
    // Forbid copy constructor.
    StManColumnKarma (const StManColumnKarma&);

    // Forbid assignment.
    StManColumnKarma& operator= (const StManColumnKarma&);

    // Get a pointer to the scalar value in the given row.
    // Karma knows its data type.
    const void* getDataPtr (uInt rownr);

    // Put the scalar value into the given row.
    // Karma knows its data type.
    void putDataPtr (uInt rownr, const void* dataPtr);
};




//*display 4
// 
// <summary> Karma table storage manager class </summary>
// <reviewed reviewer="" date="" tests="">
//
// <prerequisite>
//   <li> StManColumn
//   <li> StManKarma
// </prerequisite>
//
// <etymology>
// StManKarma is the storage manager using Karma.
// </etymology>
//
// <synopsis> 
// StManKarma is a table storage manager based on Karma.
// It uses the Karma storage package to read and write data in
// table columns. Currently this package is memory based, but
// in the (near?) future it will also support regular IO.
//
// Only columns containing scalars or direct arrays are supported.
// All standard data types, except Strings, are supported.
// In Karma arrays are stored in C-order, while the AIPS++ arrays
// are in Fortran order. This means that the axes order is reversed.
// Indirect arrays may be supported in the further future.
// Addition and removal of rows and columns is not supported.
// The exact number of rows have to be allocated at table creation time.
//
// The data of this storage manager is kept in one file. Its name is
// the table name appended with .N_Karma.kf, where N is the (unique)
// storage manager sequence number.
// </synopsis> 
//
// <motivation>
// </motivation>
//
// <todo asof="$DATE:$">
// </todo>

class StManKarma : public StMan
{
public:

    // Create a storage manager for the given new table.
    StManKarma (SetupNewTable&);

    ~StManKarma();

    // Get the id of this class (i.e. StManKarma).
    String id() const;

    // Get the type of storage manager (i.e. Karma).
    String type() const;

    // Get the nr of rows in this storage manager .
    uInt nrow() const
	{ return nrrow_p; }

    // Get the Karma data store.
    DataStore dataStore()
	{ return dataStore_p; }

    // Get the Karma row coordinate name.
    const String& rowCoordName() const
	{ return rowCoordName_p; }

    // Create a column in the storage manager.
    // This is not possible for data type TpString.
    // <group>
    // Create a scalar column.
    StManColumn* createScalarColumn (const String& name, int dataType);
    // Create a direct array column.
    StManColumn* createDirArrColumn (const String& name, int dataType);
    // Create an indirect array column (not possible).
    StManColumn* createIndArrColumn (const String& name, int dataType);
    // </group>

private:
    // The number of rows in the columns.
    uInt nrrow_p;
    // The number of scalar columns.
    uInt nrscalar_p;
    // The assembly of all columns (actually Block<StManColumnKarma*>).
    Block<StManColumnKarma*>  colSet_p;
//#//    Block<void*> colSet_p;
    // The Karma columns description.
    packet_desc* karmaDesc_p;
    // The Karma data store file.
    DataStore    dataStore_p;
    // The rows scalar coordinate name for Karma (= SCALARS_ROWNR).
    String       rowCoordName_p;

    // Flush the data and close the storage manager..
    void close();

    // Handle the creation of the storage manager for a new table.
    // This creates the necessary Karma data structures.
    void doCreate (uInt nrrow);

    // Open the storage manager for an existing table.
    void open (uInt nrrow);
};



#endif
