//# Table.h: Main interface classes to tables
//# Copyright (C) 1994
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: aips2-request@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id: Table.h,v 1.1.1.1 1996/04/04 16:37:10 dschieb Exp $

#if !defined(AIPS_TABLE_H)
#define AIPS_TABLE_H

#if defined(_AIX)
#pragma implementation ("../../implement/Table/Table.cc")
#endif 

//# Includes
#include <aips/aips.h>
#include <aips/Tables/BaseTable.h>
#include <aips/Misc/Sort.h>

//# Forward Declarations
class SetupNewTable;
class TableDesc;
class ColumnDesc;
class TableKeywordSet;
class TableExprNode;
template<class T> class Vector;
template<class T> class Block;


// 
// <summary> Main interface class to a read/write table </summary>
// <reviewed reviewer="TPPR" date="08.11.94" tests="none">

// <prerequisite>
//   <li> SetupNewTable
//   <li> TableDesc
//   <li> TableColumn 
//   <li> ScalarColumn
//   <li> ArrayColumn
// </prerequisite>

// <synopsis>
// Class Table can be used to create a new table or to access an existing
// table in read/write or readonly mode.
//
// To access the data in a Table, objects have to be created
// to access the columns. These objects are TableColumn,
// ScalarColumn<T> and ArrayColumn<T>, which can be created
// via their constructors.
// Furthermore the Table has a TableKeywordSet object
// which can be read or written using the appropriate functions.
//
// To open an existing table, a simple Table constructor can be used.
// The possible construct options are:
// <ul>
//   <li> Old            readonly table (default option)
//   <li> Update         update existing table
//   <li> Delete         delete table
// </ul>
// Creating a new table requires more work, because columns have
// to be bound to storage managers or virtual column engines.
// Class SetupNewTable is needed for this purpose. The Tables module
// documentation explains in more detail how to create a table.
//
// Other Table objects can be created from a Table using
// the select, project and sort functions. In that way a subset
// of the table can be created and it can be read/written in the same
// way as a normal Table. However, writing has the effect that the
// underlying table gets written.
//
// See the <linkto module=Tables>Tables</linkto> module for more information.
// </synopsis>

// <example>
// <srcblock>
// // Open a table to be updated.
// Table myTable ("theTable", Table::Update);
// // Write the column containing the scalar RA.
// ScalarColumn<double> raColumn(myTable, "RA");
// uInt nrrow = myTable.nrow();
// for (uInt i=0; i<nrrow; i++) {
//    raColumn.put (i, i+10);    // Put value i+10 into row i
// }
// </srcblock>
// </example>

// <motivation>
// Table is the envelope for the underlying counted referenced
// classes derived from BaseTable. In this way no pointers have
// to be used to get polymorphism.
// </motivation>

// <todo asof="$DATE:$">
//   <li> add, remove, rename columns.
//   <li> virtual concatenation of tables (if still necessary).
//   <li> maybe an isAttached function.
// </todo>

class Table : public Cleanup
{
friend class ROTableColumn;
friend class PlainTable;
friend class RefTable;
friend class TableIterator;

public:
    enum TableOption {Old=1,                 // existing table
		      New,                   // create table
		      NewNoReplace,          // create table (may not exist)
		      Scratch,               // scratch table
		      Update,                // update existing table
		      Delete};               // delete table

    // Create a null Table object (i.e. no table is attached yet).
    // The sole purpose of this constructor is to allow construction
    // of an array of Table objects.
    // The assignment operator can be used to make a null object
    // reference a column.
    // Note that sort functions, etc. will cause a segmentation fault
    // when operating on a null object. It was felt it was too expensive
    // to test on null over and over again. The user should use the isNull
    // or throwIfNull function in case of doubt.
    Table();

    // Create a table object for an existing writable table.
    // The only options allowed are Old, Update, and Delete.
    // <group>
    Table (const String& tableName, TableOption = Table::Old);
    // Check if the table description has the given name.
    Table (const String& tableName, const String& tableDescName,
	   TableOption = Table::Old);
    // </group>

    // Make a table object for a new table, which can thereafter be used
    // for reading and writing.
    // If there are unbound columns, default storage managers an/ord virtual
    // column engines will be created and bound to those columns.
    // Create the table with the given nr of rows. If a storage manager
    // is used which does not allow addition of rows, the number of rows
    // in the table must already be given here.
    // Optionally the rows can be initialized with the default
    // values as defined in the column descriptions.
    Table (SetupNewTable&, uInt nrrow = 0, Bool initialize = False);

    //# Virtually concatenate some tables.
    //# All tables must have the same description.
//#//    Table (const Block<Table>&);

    // Copy constructor (reference semantics).
    Table (const Table&);

    // The destructor writes the table if necessary.
    ~Table();

    //*display 8
    // This function is used by the exception handling mechanism we have
    // defined. It merely calls the destructor. When real exceptions are
    // available it will be unnecessary.
    void cleanup();

    // Assignment (reference semantics).
    Table& operator= (const Table&);

    // Test if the object is null, i.e. does not reference a table yet.
    // This is the case if the default constructor is used.
    Bool isNull() const
	{ return (baseTabPtr_p == 0  ?  True : False); }

    // Throw an exception if the object is null, i.e.
    // if function isNull() is True.
    void throwIfNull() const;

    // Write a reference table (i.e. the result of a sort or select)
    // into a file with the given name.
    // This can be read back as a normal table.
    void writeRefTable (const String& fileName) const;

    // Test if a table file with the given name exists.
    static Bool fileExists (const String& tableFileName);

    // Test if the table file with the given name is writable.
    // It throws an exception if the table does not exist.
    static Bool fileIsWritable (const String& tableFileName);

    // Test if this table is writable.
    Bool isWritable() const;

    // Test if the given column is writable.
    // <group>
    Bool isWritable (const String& columnName) const;
    Bool isWritable (uInt columnIndex) const;
    // </group>

    // Get access to the table keyword set.
    const TableKeywordSet& keywordSet() const;
    TableKeywordSet& keywordSet();

    // Get access to the table description.
    // This can be used to get nr of columns, etc..
    // <note role=tip> The keyword set in the table description is not the
    // same set as the keyword set in the table itself.
    const TableDesc& tableDesc() const;

    // Get the table name.
    const String& tableName() const;

    // Get the table option.
    int tableOption() const;

    // Get nr of rows.
    uInt nrow() const;

    // Test if it is possible to add a row to this table.
    // It is possible if all storage managers used for the table
    // support it.
    Bool canAddRow() const;

    // Add one or more rows at the end of the table.
    // This will fail for tables not supporting addition of rows.
    // Optionally the rows can be initialized with the default
    // values as defined in the column descriptions.
    void addRow (uInt nrrow = 1, Bool initialize = False);

    // Test if it is possible to remove a row from this table.
    // It is possible if all storage managers used for the table
    // support it.
    Bool canRemoveRow() const;

    // Remove rows.
    // This will fail for tables not supporting removal of rows.
    // <group>
    // Remove the given row.
    void removeRow (uInt rownr);

    // Remove multiple rows.
    // This can be useful with the select and rowNumbers functions
    // to remove some selected rows from the table.
    void removeRow (const Vector<uInt>& rownrs);
    // </group>

    // Select rows from a table using an select expression consisting
    // of TableExprNode objects. For example:
    // <srcblock>
    //    Table result = tab(tab.col("columnName") > 10);
    // </srcblock>
    // All rows for which the expression is true will be selected and
    // "stored" in the result.
    // You need to include ExprNode.h for this purpose.
    // <group>
    // Create a TableExprNode object for a keyword in the table keyword set.
    TableExprNode& key (const String& keywordName) const;
    // Create a TableExprNode object for a column
    TableExprNode& col (const String& columnName) const;
    // It will be determined if the name is a keyword or column.
    TableExprNode& keyCol (const String& name) const;

    // Select rows using the given table select expression.
    Table operator() (TableExprNode&) const;
    // </group>

    // Select rows using a vector of row numbers.
    // This can, for instance, be used to select the same rows as
    // were selected in another table (using the rowNumbers function).
    // <srcblock>
    //     Table result = thisTable (otherTable.rowNumbers());
    // </srcblock>
    Table operator() (const Vector<uInt>& rownrs) const;

    // Select rows using a mask block.
    // The length of the block must match the number of rows in the table.
    // If an element in the mask is True, the corresponding row will be
    // selected.
    Table operator() (const Block<Bool>& mask) const;

    // Project the given columns (i.e. select the columns).
    Table project (const Block<String>& columnNames) const;

    //# Virtually concatenate all tables in this column.
    //# The column cells must contain tables with the same description.
//#//    Table concatenate (const String& columnName) const;

    // Do logical operations on a table.
    // It can be used for row-selected or projected (i.e. column-selected)
    // tables. The tables involved must come from the same root table or
    // be the root table themselves.
    // <group>
    // Intersection with another table.
    Table operator& (const Table&) const;
    // Union with another table.
    Table operator| (const Table&) const;
    // Subtract another table.
    Table operator- (const Table&) const;
    // Xor with another table.
    Table operator^ (const Table&) const;
    // Take complement.
    Table operator! () const;
    // </group>

    // Sort a table on one or more columns of scalars.
    // Per column a compare function can be provided. By default
    // the standard compare function defined in Compare.h will be used.
    // Default sort order is ascending.
    // Default sorting algorithm is the heap sort.
    // <group>
    // Sort on one column.
    Table sort (const String& columnName,
		int = Sort::Ascending,
		int = Sort::HeapSort) const;
    // Sort on multiple columns. The principal column has to be the
    // first element in the Block of column names.
    // <group>
    Table sort (const Block<String>& columnNames,
		int = Sort::Ascending,
		int = Sort::HeapSort) const;
    // You can give the order per column.
    Table sort (const Block<String>& columnNames,
		const Block<Int>& sortOrders,
		int = Sort::HeapSort) const;
    // Provide some special compare functions via a function pointer.
    // A zero function pointer means using the standard compare function
    // from Compare.h.
    Table sort (const Block<String>& columnNames,
		const Block<ObjCompareFunc*>& compareFunctionPointers,
		const Block<Int>& sortOrders,
		int = Sort::HeapSort) const;
    // </group>
    // </group>

    // Get a vector of row numbers.
    // In case the table is a subset of the root table, this tells which
    // rows of the root table are part of the subset.
    // In case the table is the root table itself, the result is a vector
    // containing the row numbers 0 .. #rows-1.
    Vector<uInt> rowNumbers() const;

    // Test if a column can be added.
    Bool canAddColumn() const;

    // Add a column to the table.
    void addColumn (const ColumnDesc&);

    // Test if a column can be removed.
    Bool canRemoveColumn() const;

    // Remove a column.
    void removeColumn (const String& columnName);

    // Test if a column can be renamed.
    Bool canRenameColumn() const;

    // Rename a column.
    void renameColumn (const String& newName, const String& oldName);

    //*display 4
    // Write a table to AipsIO (for TypedKeywords<Table>).
    // This will only write the table name.
    friend AipsIO& operator<< (AipsIO&, const Table&);

    //*display 4
    // Read a table from AipsIO (for TypedKeywords<Table>).
    // This will read the table name and open the table as writable
    // if the table file is writable, otherwise as readonly.
    friend AipsIO& operator>> (AipsIO&, Table&);

    //*display 4
    // Read a table from AipsIO (for TableKeywords).
    // This will read the table name and open the table as writable
    // if the switch is set and if the table file is writable.
    // otherwise it is opened as readonly.
    void getTableKeyword (AipsIO&, Bool openWritable);

    // Write a table to ostream (for TypedKeywords<Table>).
    // This only shows its name and number of columns and rows.
    friend ostream& operator<< (ostream&, const Table&);

protected:
    BaseTable* baseTabPtr_p;                   // ptr to table representation
    //# The isCounted_p flag is normally true.
    //# Only for internally used Table objects (i.e. in the DataManager)
    //# this flag is False, otherwise a mutual dependency would exist.
    //# The DataManager has a Table object, which gets deleted by the
    //# DataManager destructor. The DataManager gets deleted by the
    //# PlainTable destructor, which gets called when the last Table
    //# object gets destructed. That would never be the case if this
    //# internally used Table object was counted.
    Bool       isCounted_p;

    // Construct a Table object from a BaseTable*.
    // By default the object gets counted.
    Table (BaseTable*, Bool countIt = True);

    // Open an existing table.
    void open (const String& nam, const String& type, int tableOption);

private:
    // Get the pointer to the underlying BaseTable.
    // This is needed for some friend classes.
    BaseTable* baseTablePtr() const;

    // Look in the cache if the table is already open.
    // If so, check if table option matches.
    BaseTable* lookCache (const String& name, int tableOption);
};



inline Bool Table::isWritable() const
    { return baseTabPtr_p->isWritable(); }
inline Bool Table::isWritable (const String& columnName) const
    { return baseTabPtr_p->isWritable (columnName); }
inline Bool Table::isWritable (uInt columnIndex) const
    { return baseTabPtr_p->isWritable (columnIndex); }

inline uInt Table::nrow() const
    { return baseTabPtr_p->nrow(); }
inline BaseTable* Table::baseTablePtr() const
    { return baseTabPtr_p; }
inline const TableDesc& Table::tableDesc() const
    { return baseTabPtr_p->tableDesc(); }
inline const TableKeywordSet& Table::keywordSet() const
    { return baseTabPtr_p->keywordSet(); }
inline TableKeywordSet& Table::keywordSet()
    { return baseTabPtr_p->keywordSet(); }
inline const String& Table::tableName() const
    { return baseTabPtr_p->tableName(); }
inline int Table::tableOption() const
    { return baseTabPtr_p->tableOption(); }

inline Bool Table::canAddRow() const
    { return baseTabPtr_p->canAddRow(); }
inline Bool Table::canRemoveRow() const
    { return baseTabPtr_p->canRemoveRow(); }
inline Bool Table::canAddColumn() const
    { return baseTabPtr_p->canAddColumn(); }
inline Bool Table::canRemoveColumn() const
    { return baseTabPtr_p->canRemoveColumn(); }
inline Bool Table::canRenameColumn() const
    { return baseTabPtr_p->canRenameColumn(); }

inline void Table::addRow (uInt nrrow, Bool initialize)
    { baseTabPtr_p->addRow (nrrow, initialize); }
inline void Table::removeRow (uInt rownr)
    { baseTabPtr_p->removeRow (rownr); }
inline void Table::removeRow (const Vector<uInt>& rownrs)
    { baseTabPtr_p->removeRow (rownrs); }
inline void Table::addColumn (const ColumnDesc& colDesc)
    { baseTabPtr_p->addColumn (colDesc); }
inline void Table::removeColumn (const String& columnName)
    { baseTabPtr_p->removeColumn (columnName); }
inline void Table::renameColumn (const String& newName, const String& oldName)
    { baseTabPtr_p->renameColumn (newName, oldName); }


#endif
