//# TableDesc.h:  specify structure of aips++ tables
//# Copyright (C) 1994
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: aips2-request@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id: TableDesc.h,v 1.1.1.1 1996/04/04 16:37:06 dschieb Exp $

#if !defined(AIPS_TABLEDESC_H)
#define AIPS_TABLEDESC_H

#if defined(_AIX)
#pragma implementation ("../../implement/Tables/TableDesc.cc")
#endif 

//# Includes
#include <aips/aips.h>
#include <aips/TabPath.h>
#include <aips/Tables/ColDescSet.h>
#include <aips/AipsIO.h>

//# Forward Declarations
class ScalarKeywordSet;

// 
// <summary> specify the structure of an AIPS++ table  </summary>
// <reviewed reviewer="Paul Shannon" date="1994/08/11" tests="none">

// <prerequisite>
//   <li> column description classes
//   <li> ScalarKeywordSet
// </prerequisite>

// <synopsis> 
// A TableDesc object contains the description, or structure, of a table.
// This description is required for the creation of a new table.  
// Descriptions are subsequently associated with every table and
// embedded in them.
//
// <a name="important">A table description consists of the following items</a>:
// <ul>
//  <li> Name, which cannot be blank if the description is saved in a file.
//         The file name will be this name followed by .tabdsc.
//  <li> Version, which defaults to a blank string.
//         It serves merely as information for the user.
//  <li> Comment, which defaults to an empty string.
//         This serves purely as an informational string for the user.
//  <li> A set of column descriptions which has to be added to the
//         table description. A column description can be created using
//         the classes ScalarColumnDesc, etc..
//         At table creation it is determined by the user if a column
//         has to be stored using a storage manager or calculated
//         on-the-fly using a so-called virtual column engine.
//  <li> A keyword set, which is by default empty.
//         When a table is created from the description, it gets
//         a copy of this keyword set as its initial keyword set.
// </ul>
//
// A TableDesc object can be constructed with one of the following
// options:
// <ul>
//  <li> Old
//    Open an existing table description file as readonly.
//  <li> Update
//    Open an existing table description file as read/write
//    The TableDesc destructor will rewrite the possibly changed
//    description.
//  <li> New
//    Create a new table description file.
//    The TableDesc destructor will write the table description into the file.
//  <li> NewNoReplace
//    As option New, but an exception will be thrown if the table
//    description file already exists.
//  <li> Scratch
//    Create a temporary table description. The table description will
//    be lost when the TableDesc object is destructed.
//    This is useful to create a Table object without storing the
//    description separately.
//    Note that the Table object maintains its own description (i.e. it
//    copies the description when being constructed).
//  <li> Delete
//    Delete the table description file. This gets done by the destructor.
// </ul>
//
// More information is provided in the Tables module documentation.
// </synopsis>
// 
// <example>
//     // First build the new description of a subtable.
//     // Define keyword subkey (integer) having value 10.
//     // Define columns ra and dec (double).
//     TableDesc subTableDesc("tTableDesc_sub", "1", TableDesc::New);
//     subTableDesc.addColumn (ScalarColumnDesc<double>("ra"));
//     subTableDesc.addColumn (ScalarColumnDesc<double>("dec"));
//
//     // Now create a new table description
//     // Define a comment for the table description.
//     // Define some keywords.
//     ColumnDesc colDesc1, colDesc2;
//     TableDesc td("tTableDesc", "1", TableDesc::New);
//     td.comment() = "A test of class TableDesc";
//     td.keywordSet().keysdouble()("equinox") = 1950;
//
//     // Define an integer column ab using the TableDesc::addColumn
//     // function which creates a scalar column description.
//     td.addColumn (ScalarColumnDesc<Int>("ab", "Comment for column ab"));
//
//     // Add a scalar integer column ac, define keywords for it
//     // and define a default value 0.
//     // Overwrite the value of keyword unit.
//     ScalarColumnDesc<Int> acColumn("ac");
//     acColumn.keywordSet().keysComplex()("scale") = 0;
//     acColumn.keywordSet().keysString()("unit") = "";
//     acColumn.setDefault (0);
//     td.addColumn (acColumn);
//     td["ac"].keywordSet().keysString()("unit") = "DEG";
//
//     // Add a scalar string column ad and define its comment string.
//     td.addColumn (ScalarColumnDesc<String>("ad","comment for ad"));
//
//     // Now define array columns.
//     // This one is indirect and has no dimensionality mentioned yet.
//     td.addColumn (ArrayColumnDesc<Complex>("Arr1","comment for Arr1"));
//     // This one is indirect and has 3-dim arrays.
//     td.addColumn (ArrayColumnDesc<Int>("A2r1","comment for Arr1",3));
//     // This one is direct and has 2-dim arrays with axes length 4 and 7.
//     td.addColumn (ArrayColumnDesc<uInt>("Arr3","comment for Arr1",
//        				   IPosition(2,4,7),
// 					   ColumnDesc::Direct));
//
//     // Add a columns containing tables.
//     td.addColumn (SubTableDesc("sub1", "subtable by name","tTableDesc_sub"));
// }
// </example>

// <motivation>
// A table description specifies the structure, but not the contents,
// of an aips++ table.  Since many tables will have identical structure
// and different content, it makes good sense to separate structure 
// ("description") from content.
// </motivation>
//
// <todo asof="$DATE:$">
// </todo>

class TableDesc
{
public:

    //# Enumerate the possible options for TableDesc.
    enum TDOption {Old=1, New, NewNoReplace, Scratch, Update, Delete};

    // The default constructor creates a table description with
    // option = Scratch and a blank name.
    TableDesc();

    // Create a table description object with the given name.
    // This name can be seen as the table type in the same way as a
    // class name is the data type of an object.
    // The name can only be blank when option=Scratch.
    // The default table description path is used for the description file.
    TableDesc (const String& type, TDOption = Old);

    // Create a table description object with the given name (i.e. table type)
    // and version.
    // The name can only be blank when option=Scratch.
    // The default table description path is used for the description file.
    TableDesc (const String& type, const String& version, TDOption = Old);

    // Create a table description object.
    // The given table description path is used for the description file.
    // The name can only be blank with option=Scratch.
    TableDesc (const String& type, const String& version,
	       const TabPath&, TDOption = Old);

    // Create a table description object with the given name (i.e. table type)
    // and version by copying the input table description.
    // If the given name or version is blank, it will be copied from
    // the input table description.
    // The default table description path is used for the description file.
    // The only options allowed are New, NewNoReplace and Scratch.
    TableDesc (const TableDesc&, const String& type, const String& version,
	       TDOption);

    // Create a table description object with the given name (i.e. table type)
    // and version by copying the input table description.
    // If the given name or version is blank, it will be copied from
    // the input table description.
    // The given table description path is used for the description file.
    // The only options allowed are New, NewNoReplace and Scratch.
    TableDesc (const TableDesc&, const String& type, const String& version,
	       const TabPath&, TDOption);

    // This copy constructor makes a copy of the table description
    // maintaining its name and version. By default a Scratch copy is made.
    // It serves as a shorthand for the constructor:
    // <srcblock>
    //    TableDesc (const TableDesc&, "", "", TDOption);
    // </srcblock>
    TableDesc (const TableDesc&, TDOption = Scratch);

    // The destructor writes the table description if changed.
    ~TableDesc();

    // Get access to the set of column descriptions.
    // This can be used to add it to another TableDesc (see function add).
    // ColumnDescSet is derived from KeywordSet, so all const
    // KeywordSet functions (merge, etc.) can be used (e.g. isEqual to
    // test if 2 table descriptions are the same).
    const ColumnDescSet& columnDescSet() const;

    // Add an existing column set to this description.
    // The two column sets have to be disjoint.
    // <thrown>
    //   <li> DuplicateKeyword
    // </thrown>
    void add (const ColumnDescSet&);

    // Get access to the keyword set.
    // <group>
    ScalarKeywordSet& keywordSet();
    const ScalarKeywordSet& keywordSet() const;
    // </group>

    // Add a column to the table description.
    // An exception is thrown if a keyword or column with this name
    // already exists.
    // Although this function has a ColumnDesc as argument, it is usually
    // needed to construct a more specialized object like
    // ArrayColumnDesc<float>. A ColumnDesc constructor converts
    // that automatically to a ColumnDesc object.
    // <srcblock>
    //   tableDesc.addColumn (ArrayColumnDesc<float> ("NAME"));
    // </srcblock>
    // On the other hand this function can also be used to add a
    // column description from another table as in:
    // <srcblock>
    //   tableDesc.addColumn (otherTableDesc.columnDesc("NAME"));
    // </srcblock>
    // <group>
    ColumnDesc& addColumn (const ColumnDesc&);

    // Add a column to the table description and rename it.
    // This may be useful to use a description of another column.
    ColumnDesc& addColumn (const ColumnDesc&, const String& newname);
    // </group>

    // Remove a column.
    // An exception is thrown if the column does not exist.
    void removeColumn (const String& name);

    // Rename a column.
    // An exception is thrown if the old name does not exist or
    // if the name already exists.
    void renameColumn (const String& newname, const String& oldname);

    // Get number of columns.
    Int ncolumn() const;

    // Test if a column with this name exists.
    Bool isColumn (const String& name) const;

    // Get the column description by name or by index.
    // An exception is thrown if the column does not exist.
    // Function isColumn should be used to test if a column exists.
    // <group>
    ColumnDesc& columnDesc (const String& name);
    const ColumnDesc& columnDesc (const String& name) const;
    ColumnDesc& operator[] (const String& name);
    const ColumnDesc& operator[] (const String& name) const;
    ColumnDesc& columnDesc (uInt index);
    const ColumnDesc& columnDesc (uInt index) const;
    ColumnDesc& operator[] (uInt index);
    const ColumnDesc& operator[] (uInt index) const;
    // </group>

    // Get comment string.
    const String& comment() const;

    // Get comment string (allowing it to be changed).
    String& comment();

    // Show the table description.
    void show() const;

    // Get the table type (i.e. name of table description).
    const String& getType() const;

    // Get the table description version.
    const String& version() const;

    //*display 4
    // Check recursively if the descriptions of all subtables are known.
    void checkSubTableDesc() const;

private:
    String             name_p;              // name of table description
    String             vers_p;              // version of table description
    String             dir_p;               // directory
    String             comm_p;              // comment
    ScalarKeywordSet*  key_p;               // set of keywords
    ColumnDescSet      col_p;               // set of column names + indices
    Bool               swwrite_p;           // True = description can be written
    TDOption           option_p;            // Table desc. open option
    AipsIO             iofil_p;             // File

    // Assignment is not supported, because it is impossible to define
    // its semantics. Does the data need to be written into a file
    // before being overwritten?
    // Declaring it private, makes it unusable.
    TableDesc& operator= (const TableDesc&);

    // Initialize the table description.
    void init (const TabPath&);

    // Initialize and copy a table description.
    void copy (const TableDesc&, const TabPath&);

public:
//*display 4
// This function is not really meant for public use,
// but only for other Table classes.
// <group>

    // Put the table description into the file.
    // The flag determines if descriptions of nested tables will
    // also be written.
    //   False= write only nested desc. with name==""
    //          This is for saving a description in a file. In that way
    //          descriptions referred via a name are not saved with this
    //          description.
    //   True = write all nested descriptions.
    //          This is for writing a table, which should contain the
    //          descriptions of all nested tables.
    void putFile (AipsIO&, Bool) const;

    // Get the table description from the file.
    void getFile (AipsIO&);

// </group>

};


//# Get number of columns.
inline Int TableDesc::ncolumn () const
    { return col_p.nkey(); }

//# Test if column exists.
inline Bool TableDesc::isColumn (const String& name) const
    { return col_p.isDefined(name); }

//# Get a column description.
inline ColumnDesc& TableDesc::columnDesc (const String& name)
    { return col_p[name]; }
inline const ColumnDesc& TableDesc::columnDesc (const String& name) const
    { return col_p[name]; }
inline ColumnDesc& TableDesc::operator[] (const String& name)
    { return col_p[name]; }
inline const ColumnDesc& TableDesc::operator[] (const String& name) const
    { return col_p[name]; }
inline ColumnDesc& TableDesc::columnDesc (uInt index)
    { return col_p[index]; }
inline const ColumnDesc& TableDesc::columnDesc (uInt index) const
    { return col_p[index]; }
inline ColumnDesc& TableDesc::operator[] (uInt index)
    { return col_p[index]; }
inline const ColumnDesc& TableDesc::operator[] (uInt index) const
    { return col_p[index]; }


//# Return the name (ie. type) of the table description.
inline const String& TableDesc::getType () const
    { return name_p; }

//# Return the version of the table description.
inline const String& TableDesc::version () const
    { return vers_p; }

//# Get access to the set of keywords.
inline ScalarKeywordSet& TableDesc::keywordSet ()
    { return *key_p; }
inline const ScalarKeywordSet& TableDesc::keywordSet () const
    { return *key_p; }

//# Get the set of columns.
inline const ColumnDescSet& TableDesc::columnDescSet() const
    { return col_p; }

//# Add another set of columns.
inline void TableDesc::add (const ColumnDescSet& set)
    { col_p.add (set); }

//# Add a column.
inline ColumnDesc& TableDesc::addColumn (const ColumnDesc& column)
    { return col_p.addColumn (column); }

inline ColumnDesc& TableDesc::addColumn (const ColumnDesc& column,
					 const String& newname)
    { return col_p.addColumn (column, newname); }

//# Remove a column.
inline void TableDesc::removeColumn (const String& name)
    { col_p.remove (name); }

//# Rename a column.
inline void TableDesc::renameColumn (const String& newname,
					    const String& oldname)
    { col_p.rename (newname, oldname); }

//# Access the comment.
inline const String& TableDesc::comment () const
    { return comm_p; }

inline String& TableDesc::comment ()
    { return comm_p; }

inline void TableDesc::checkSubTableDesc () const
   { col_p.checkSubTableDesc(); }



extern TabPath tabdescpath;                  // default TabDesc search path


#endif

