//# TableKeywords.h: Keyword values representing tables
//# Copyright (C) 1994
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: aips2-request@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id: TableKeywords.h,v 1.1.1.1 1996/04/04 16:37:11 dschieb Exp $

#if !defined(AIPS_TABLEKEYWORDS_H)
#define AIPS_TABLEKEYWORDS_H

#if defined(_AIX)
#pragma implementation ("../../implement/Tables/TableKeywords.cc")
#endif

//# Includes
#include <aips/aips.h>
#include <aips/TypedKeywords.h>
#include <aips/Tables/Table.h>

//# Forward Declarations


// 
// <summary>  Keyword values representing tables </summary>
// <reviewed reviewer="Gareth Hunt" date="94Nov17" tests="">
//
// <prerequisite>
//   <li> Table
//   <li> TableKeywordSet
//   <li> TypedKeywords
// </prerequisite>
//
// <synopsis>
// TableKeywords represents the set of keywords containing tables.
// It is derived from TypedKeywords<Table> and adds one thing.
// When reading the table back, a switch is passed to tell if
// the table should be opened as read/write or as readonly.
// In this way a table can force that the tables in its keywords
// are opened in the same way as the table itself.
//
// In the future possible other things could be done:
// <ul>
//   <li> Open a table when needed, thus when the
//          keyword is accessed for the first time.
//   <li> When a table gets written, it is possible to write it as
//          part of the parent table.
// </ul>
//
// To the user this class is identical to TypedKeywords<Table>.
// All TypedKeywords functions can be used in the standard way.

// <motivation>
// This class provides the extra functionality for keywords containing
// tables. This is needed because tables are much more complex entities
// than scalars or arrays.
// </motivation>
//
// <example>
// <srcblock>
// // Open the table and get access to the table keyword set.
// // Use that to get access to the keywords containing tables.
// Table table("table.data");
// const TableKeywordSet& keyset = table.keywordSet();
// const TableKeywords& tkeys = keyset.keysTable();
// // Now get the table assigned to keyword keytab.
// Table keyTab = tkeys("keytab");
// // Of course it could be combined into one statement:
// //     Table keyTab = table.keywordSet().keysTable()("keytab");
// </srcblock>
// </example>
//
// <todo asof="$DATE:$">
//   <li> allow writing into parent table
//          (this could be done when the keyword table is a scratch table).
// </todo>


class TableKeywords : public TypedKeywords<Table>
{
public:

    // Construct with a given name pool and the data type used.
    // The default value is needed for the SimpleOrderedMap constructor.
    TableKeywords (KeywordSet* pool, DataType, const Table& dflt);

    ~TableKeywords();

    // Read from AipsIO stream.
    void getData (AipsIO& ios, Bool openWritable);

private:
    // Forbid copy constructor.
    TableKeywords (const TableKeywords&);

    // Forbid assignment.
    TableKeywords& operator= (const TableKeywords&);
};


#endif
